/*

The contents of this web application are subject to the Mozilla Public License Version 1.1 (the "License"); you may not use this web application except in compliance with the License. 
You may obtain a copy of the License at http://www.mozilla.org/MPL/.

Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. 
See the License for the specific language governing rights and limitations under the License.

The Original Code is owned by and the Initial Developer of the Original Code is Composite A/S (Danish business reg.no. 21744409). All Rights Reserved

Section 11 of the License is EXPRESSLY amended to include a provision stating that any dispute, including but not limited to disputes related to the enforcement of the License, to which Composite A/S as owner of the Original Code, as Initial Developer or in any other role, becomes a part to shall be governed by Danish law and be initiated before the Copenhagen City Court ("K�benhavns Byret")
                        
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.UI;
using System.Workflow.Runtime;
using System.Xml.Linq;
using System.Xml.Serialization;
using Composite.C1Console.Actions;
using Composite.C1Console.Events;
using Composite.Data;
using Composite.Data.Types;
using Composite.Functions;
using Composite.Functions.ManagedParameters;
using Composite.Core.Logging;
using Composite.Plugins.Functions.FunctionProviders.SqlFunctionProvider;
using Composite.Core.WebClient;
using Composite.Core.WebClient.FlowMediators.FormFlowRendering;
using Composite.Core.WebClient.FunctionCallEditor;
using Composite.Core.WebClient.State;
using Composite.C1Console.Workflow;
using Composite.C1Console.Workflow.Foundation;


namespace Composite.Plugins.Elements.ElementProviders.SqlFunctionElementProvider
{
    [EntityTokenLock()]
    [AllowPersistingWorkflow(WorkflowPersistingType.Idle)]
    public sealed partial class EditSqlFunctionProviderWorkflow : Composite.C1Console.Workflow.Activities.FormsWorkflow
    {
        public EditSqlFunctionProviderWorkflow()
        {
            InitializeComponent();
        }



        private void initialize_ExecuteCode(object sender, EventArgs e)
        {
            DataEntityToken token = this.EntityToken as DataEntityToken;

            ISqlFunctionInfo functionInfo = (ISqlFunctionInfo)token.Data;
            IEnumerable<ManagedParameterDefinition> parameters = ManagedParameterManager.Load(functionInfo.Id);

            this.Bindings.Add("SqlQuery", functionInfo);
            this.Bindings.Add("Parameters", parameters);
            this.Bindings.Add("ParameterTypeOptions", GetParameterTypes().ToList());

            // Creating a session state object
            Guid stateId = Guid.NewGuid();
            var state = new EditSqlFunctionState { WorkflowId = WorkflowInstanceId, ConsoleIdInternal = GetCurrentConsoleId() };
            SessionStateManager.DefaultProvider.AddState<IParameterEditorState>(stateId, state, DateTime.Now.AddDays(7.0));

            this.Bindings.Add("SessionStateProvider", SessionStateManager.DefaultProviderName);
            this.Bindings.Add("SessionStateId", stateId);
        }



        private IEnumerable<Type> GetParameterTypes()
        {
            yield return typeof(string);
            yield return typeof(int);
            yield return typeof(decimal);
            yield return typeof(DateTime);
            yield return typeof(bool);
            yield return typeof(Guid);
        }


        private void saveCodeActivity_ExecuteCode(object sender, EventArgs e)
        {
            UpdateTreeRefresher updateTreeRefresher = this.CreateUpdateTreeRefresher(this.EntityToken);
            ISqlFunctionInfo info;

            try
            {
                info = this.GetBinding<ISqlFunctionInfo>("SqlQuery");
                var parameters = this.GetBinding<IEnumerable<ManagedParameterDefinition>>("Parameters");

                ManagedParameterManager.Save(info.Id, parameters);
                DataFacade.Update(info);
            }
            catch (Exception ex)
            {
                LoggingService.LogCritical("SQL Function Save", ex);

                FlowControllerServicesContainer serviceContainer = WorkflowFacade.GetFlowControllerServicesContainer(WorkflowEnvironment.WorkflowInstanceId);
                var consoleMsgService = serviceContainer.GetService<IManagementConsoleMessageService>();
                consoleMsgService.ShowMessage(DialogType.Error, "Error", ex.Message);

                SetSaveStatus(false);
                return;
            }

            updateTreeRefresher.PostRefreshMesseges(info.GetDataEntityToken());
            SetSaveStatus(true);
        }



        private void previewCodeActivity_ExecuteCode(object sender, EventArgs e)
        {
            try
            {
                ISqlFunctionInfo queryInfo = this.GetBinding<ISqlFunctionInfo>("SqlQuery");


                var parameterDefinitions = this.GetBinding<IEnumerable<ManagedParameterDefinition>>("Parameters");
                ParameterList parameters = ManagedParameterManager.GetParametersListForTest(parameterDefinitions);
                IEnumerable<ParameterProfile> parameterProfiles = ManagedParameterManager.GetParameterProfiles(parameterDefinitions);

                SqlFunction function = new SqlFunction(queryInfo, parameterProfiles);

                XElement result = function.Execute(parameters, new FunctionContextContainer()) as XElement;

                Page currentPage = HttpContext.Current.Handler as Page;
                if (currentPage == null) throw new InvalidOperationException("The Current HttpContext Handler must be a System.Web.Ui.Page");

                UserControl inOutControl = (UserControl)currentPage.LoadControl(UrlUtils.ResolveAdminUrl("controls/Misc/MarkupInOutView.ascx"));
                inOutControl.Attributes.Add("out", result.ToString());

                FlowControllerServicesContainer serviceContainer = WorkflowFacade.GetFlowControllerServicesContainer(WorkflowEnvironment.WorkflowInstanceId);
                var webRenderService = serviceContainer.GetService<IFormFlowWebRenderingService>();
                webRenderService.SetNewPageOutput(inOutControl);

            }
            catch (Exception ex)
            {
                FlowControllerServicesContainer serviceContainer = WorkflowFacade.GetFlowControllerServicesContainer(WorkflowEnvironment.WorkflowInstanceId);
                Control errOutput = new LiteralControl("<pre>" + ex.ToString() + "</pre>");
                var webRenderService = serviceContainer.GetService<IFormFlowWebRenderingService>();
                webRenderService.SetNewPageOutput(errOutput);
            }
        }

        [Serializable]
        public sealed class EditSqlFunctionState : IParameterEditorState
        {
            public Guid WorkflowId { get; set; }
            public string ConsoleIdInternal { get; set; }

            private FormData GetFormData()
            {
                return WorkflowFacade.GetFormData(WorkflowId);
            }


            [XmlIgnore]
            public List<ManagedParameterDefinition> Parameters
            {
                get { return GetFormData().Bindings["Parameters"] as List<ManagedParameterDefinition>; }
                set { GetFormData().Bindings["Parameters"] = value; }
            }

            [XmlIgnore]
            public List<Type> ParameterTypeOptions
            {
                get { return (GetFormData().Bindings["ParameterTypeOptions"] as IEnumerable<Type>).ToList(); }
                set { GetFormData().Bindings["ParameterTypeOptions"] = value.ToList(); }
            }
        }
    }
}
