/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections;
using System.Reflection;
using System.Web;
using System.Web.WebPages;

namespace Composite.AspNet.Razor
{
    /// <summary>
    /// Replaces HttpContext.Items collection, with a custom one, that allows WebPage-s to have own execution stacks, allowing them to be executed in parallel.
    /// </summary>
    internal class CustomHttpContextWrapper: HttpContextWrapper
    {
        private readonly IDictionary _items;

        public CustomHttpContextWrapper(HttpContext context): base(context)
        {
            _items = new FixedItemsCollection(context.Items);
        }

        public override System.Collections.IDictionary Items
        {
            get
            {
                return _items;
            }
        }


        private class FixedItemsCollection: IDictionary
        {
            private readonly IDictionary _innerDictionary;
            private static readonly object _webPageStackKey;

            private object _templateStack;

            static FixedItemsCollection()
            {
                _webPageStackKey = typeof (TemplateStack).GetField("_contextKey", BindingFlags.NonPublic | BindingFlags.Static).GetValue(null);
            }

            public FixedItemsCollection(IDictionary innerDictionary)
            {
                _innerDictionary = innerDictionary;
            }

            public void Add(object key, object value)
            {
                if (key == _webPageStackKey)
                {
                    _templateStack = value;
                    return;
                }

                _innerDictionary.Add(key, value);
            }

            public void Clear()
            {
                throw new NotSupportedException();
            }

            public bool Contains(object key)
            {
                if (key == _webPageStackKey)
                {
                    return _templateStack != null;
                }

                return _innerDictionary.Contains(key);
            }

            public IDictionaryEnumerator GetEnumerator()
            {
                return _innerDictionary.GetEnumerator();
            }

            public bool IsFixedSize
            {
                get { return _innerDictionary.IsFixedSize; }
            }

            public bool IsReadOnly
            {
                get { return _innerDictionary.IsReadOnly; }
            }

            public ICollection Keys
            {
                get { return _innerDictionary.Keys; }
            }

            public void Remove(object key)
            {
                if (key == _webPageStackKey)
                {
                    _templateStack = null;
                    return;
                }

                _innerDictionary.Remove(_templateStack);
            }

            public ICollection Values
            {
                get { return _innerDictionary.Values; }
            }

            public object this[object key]
            {
                get
                {
                    if (key == _webPageStackKey)
                    {
                        return _templateStack;
                    }

                    return _innerDictionary[key];
                }
                set
                {
                    if (key == _webPageStackKey)
                    {
                        _templateStack = value;
                        return;
                    }

                    _innerDictionary[key] = value;
                }
            }

            public void CopyTo(Array array, int index)
            {
                _innerDictionary.CopyTo(array, index);
            }

            public int Count
            {
                get { return _innerDictionary.Count; }
            }

            public bool IsSynchronized
            {
                get { return _innerDictionary.IsSynchronized; }
            }

            public object SyncRoot
            {
                get { return _innerDictionary.SyncRoot; }
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return _innerDictionary.GetEnumerator();
            }
        }
    }
}
