/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Runtime.Caching;
using System.Web;
using Composite.Core.Routing;
using Composite.Data;

namespace Composite.AspNet
{
    public partial class CompositeC1SiteMapProvider
    {
        /// <summary>
        /// Wrapper for a sitemap created by <see cref="SitemapNavigator"/>
        /// </summary>
        protected class SiteMapContainer
        {
            private const string _cachePrefix = "sitemap";

            /// <exclude />
            public SiteMapNode Root { get; set; }
            /// <exclude />
            public IDictionary<string, SiteMapNode> KeyToNodesMap { get; private set; }
            /// <exclude />
            public IDictionary<string, SiteMapNode> RawUrlToNodesMap { get; private set; }
            /// <exclude />
            public IDictionary<string, SiteMapNode> ParentNodesMap { get; private set; }
            /// <exclude />
            public IDictionary<string, SiteMapNodeCollection> ChildCollectionsMap { get; private set; }

            /// <exclude />
            public SiteMapContainer()
            {
                KeyToNodesMap = new Dictionary<string, SiteMapNode>();
                RawUrlToNodesMap = new Dictionary<string, SiteMapNode>();
                ParentNodesMap = new Dictionary<string, SiteMapNode>();
                ChildCollectionsMap = new Dictionary<string, SiteMapNodeCollection>();
            }

            /// <exclude />
            public static SiteMapContainer LoadSiteMap(CompositeC1SiteMapProvider provider, CultureInfo culture, PublicationScope publicationScope, Guid rootPageId)
            {
                using (var data = new DataConnection(publicationScope, culture))
                {
                    var rootPage = data.SitemapNavigator.GetPageNodeById(rootPageId);

                    if (rootPage == null)
                    {
                        return null;
                    }

                    var container = new SiteMapContainer
                    {
                        Root = new CompositeC1SiteMapNode(provider, rootPage, data)
                    };

                    LoadNodes(provider, rootPage, null, container, data);

                    return container;
                }
            }

            private static void LoadNodes(CompositeC1SiteMapProvider provider, PageNode pageNode, SiteMapNode parent, SiteMapContainer container, DataConnection data)
            {
                if (pageNode.Url == null)
                {
                    return;
                }

                var node = new CompositeC1SiteMapNode(provider, pageNode, data);
                AddNode(node, parent, container);

                var childs = pageNode.ChildPages;
                foreach (var child in childs)
                {
                    LoadNodes(provider, child, node, container, data);
                }
            }

            private static void AddNode(SiteMapNode node, SiteMapNode parentNode, SiteMapContainer container)
            {
                Verify.ArgumentNotNull(node, "node");
                Verify.ArgumentNotNull(container, "container");

                container.KeyToNodesMap.Add(node.Key, node);
                container.ParentNodesMap.Add(node.Key, parentNode);

                if (!container.RawUrlToNodesMap.ContainsKey(node.Url))
                {
                    container.RawUrlToNodesMap.Add(node.Url, node);
                }

                if (parentNode != null)
                {
                    if (!container.ChildCollectionsMap.ContainsKey(parentNode.Key))
                    {
                        container.ChildCollectionsMap[parentNode.Key] = new SiteMapNodeCollection();
                    }

                    container.ChildCollectionsMap[parentNode.Key].Add(node);
                }
            }

            /// <summary>
            /// Gets sitemap from cache.
            /// </summary>
            /// <param name="host">The host.</param>
            /// <param name="culture">The culture.</param>
            /// <param name="rootPageId">The root page id.</param>
            /// <returns></returns>
            public static SiteMapContainer GetFromCache(string host, CultureInfo culture, Guid rootPageId)
            {
                var key = GetCacheKey(host, culture, rootPageId);

                var context = HttpContext.Current;
                var container = context.Items[key] as SiteMapContainer;

                if (container == null)
                {
                    if (!CanCache) return null;

                    container = MemoryCache.Default.Get(key) as SiteMapContainer;
                    if (container != null)
                    {
                        context.Items.Add(key, container);
                    }
                }

                return container;
            }


            /// <summary>
            /// Adds sitemap to cache.
            /// </summary>
            /// <param name="siteMap">The site map.</param>
            /// <param name="host">The host.</param>
            /// <param name="culture">The culture.</param>
            /// <param name="rootPageId">The root page id.</param>
            public static void AddToCache(SiteMapContainer siteMap, string host, CultureInfo culture, Guid rootPageId)
            {
                var key = GetCacheKey(host, culture, rootPageId);

                var context = HttpContext.Current;
                context.Items[key] = siteMap;

                if (CanCache)
                {
                    MemoryCache.Default.Add(key, siteMap, ObjectCache.InfiniteAbsoluteExpiration);
                }
            }

            private static string GetCacheKey(string host, CultureInfo culture, Guid rootPageId)
            {
                string hostnameKey;

                var urlSpace = new UrlSpace();
                if (urlSpace.ForceRelativeUrls)
                {
                    hostnameKey = string.Empty;
                }
                else
                {
                    hostnameKey = host;
                }

                return _cachePrefix + hostnameKey + rootPageId.ToString() + culture.Name + PublicationScope;
            }

            private static PublicationScope PublicationScope
            {
                get
                {
                    return DataScopeManager.CurrentDataScope.ToPublicationScope();
                }
            }

            /// <exclude />
            protected static bool CanCache
            {
                get
                {
                    return PublicationScope == PublicationScope.Published;
                }
            }

            internal static void ClearCache()
            {
                var keys = MemoryCache.Default.Where(o => o.Key.StartsWith(_cachePrefix)).Select(o => o.Key);
                foreach (var key in keys)
                {
                    MemoryCache.Default.Remove(key);
                };
            }
        }
    }
}
