/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Xml.Linq;
using Composite.C1Console.Elements.Plugins.ElementAttachingProvider;
using Composite.C1Console.Events;
using Composite.C1Console.Security;
using Composite.Core;
using Composite.Core.Extensions;
using Composite.Core.Instrumentation;


namespace Composite.C1Console.Trees
{
    /// <summary>    
    /// </summary>
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)] 
    public static class TreeFacade
    {
        private static readonly string LogTitle = "TreeFacade";

        private static readonly ITreeFacade _implementation = new TreeFacadeImpl();
        private static readonly object _lock = new object();
        private static bool _initialized;


        static TreeFacade()
        {
            GlobalEventSystemFacade.SubscribeToPostFlushEvent(OnPostFlushEvent);
        }



        /// <exclude />
        private static void EnsureInitialized()
        {
            if (_initialized) return;

            lock (_lock)
            {
                if (_initialized) return;

                using (new LogExecutionTime(LogTitle, "Initializing tree system"))
                {
                    _implementation.Initialize();
                    _initialized = true;
                }
            }
        }


        /// <summary>
        /// Returns a tree given the id of the tree or null if no tree exist with the given id
        /// </summary>
        /// <param name="treeId"></param>
        /// <returns></returns>
        public static Tree GetTree(string treeId)
        {
            EnsureInitialized();

            return _implementation.GetTree(treeId);
        }



        /// <exclude />
        public static IEnumerable<Tree> AllTrees
        {
            get
            {
                EnsureInitialized();

                return _implementation.AllTrees;
            }
        }



        /// <exclude />
        public static bool HasAttachmentPoints(EntityToken parentEntityToken)
        {
            EnsureInitialized();

            return _implementation.HasAttachmentPoints(parentEntityToken);
        }



        /// <exclude />
        public static bool HasPossibleAttachmentPoints(EntityToken parentEntityToken)
        {
            EnsureInitialized();

            return _implementation.HasPossibleAttachmentPoints(parentEntityToken);
        }



        /// <exclude />
        public static IEnumerable<Tree> GetTreesByEntityToken(EntityToken parentEntityToken)
        {
            EnsureInitialized();

            return _implementation.GetTreesByEntityToken(parentEntityToken);
        }



        /// <summary>
        /// Adds a attachment point that is persisted by the system and is loaded on every restart
        /// </summary>
        /// <param name="treeId"></param>
        /// <param name="interfaceType"></param>
        /// <param name="keyValue"></param>
        /// <param name="position"></param>
        /// <returns></returns>
        public static bool AddPersistedAttachmentPoint(string treeId, Type interfaceType, object keyValue, ElementAttachingProviderPosition position = ElementAttachingProviderPosition.Top)
        {
            EnsureInitialized();

            return _implementation.AddPersistedAttachmentPoint(treeId, interfaceType, keyValue, position);
        }



        /// <exclude />
        public static bool RemovePersistedAttachmentPoint(string treeId, Type interfaceType, object keyValue)
        {
            EnsureInitialized();

            return _implementation.RemovePersistedAttachmentPoint(treeId, interfaceType, keyValue);
        }



        /// <summary>
        /// This will add a attachment point until the system flushes.
        /// This can be used by element provider implementors to attach trees to their exising trees.
        /// </summary>
        /// <param name="treeId"></param>
        /// <param name="entityToken"></param>
        /// <param name="position"></param>
        public static bool AddCustomAttachmentPoint(string treeId, EntityToken entityToken, ElementAttachingProviderPosition position = ElementAttachingProviderPosition.Top)
        {
            EnsureInitialized();

            return _implementation.AddCustomAttachmentPoint(treeId, entityToken, position);
        }


        /// <summary>
        /// Loads a tree from an <see cref="XDocument"/>
        /// </summary>>
        /// <exclude />
        public static Tree LoadTreeFromDom(string treeId, XDocument document)
        {
            EnsureInitialized();

            return _implementation.LoadTreeFromDom(treeId, document);
        }



        private static void OnPostFlushEvent(PostFlushEventArgs args)
        {
            EnsureInitialized();

            _implementation.OnFlush();
        }
    }
}
