/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Workflow.ComponentModel;
using System.Workflow.ComponentModel.Serialization;
using System.Workflow.Runtime.Hosting;
using Composite.Core.IO;
using Composite.Core.Logging;
using Composite.Core.Types;


namespace Composite.C1Console.Workflow
{
    internal class FileWorkFlowPersisetenceService : WorkflowPersistenceService
    {
        private static readonly string LogTitle = "Workflow File Persisting Service";
        private string _baseDirectory;
        private Guid[] _abortedWorkflows;
        private readonly object _syncRoot = new object();


        public FileWorkFlowPersisetenceService(string baseDirectory)
        {
            _baseDirectory = baseDirectory;
            this.PersistAll = false;
        }



        public bool PersistAll { get; set; }

        public IEnumerable<Guid> GetAbortedWorkflows()
        {
            return _abortedWorkflows ?? new Guid[0];
        }

        public IEnumerable<Guid> GetPersistedWorkflows()
        {
            foreach (string filePath in C1Directory.GetFiles(_baseDirectory, "*.bin"))
            {
                string guidString = Path.GetFileNameWithoutExtension(filePath);

                Guid guid = Guid.Empty;
                try
                {
                    guid = new Guid(guidString);
                }
                catch { }

                if (guid.Equals(Guid.Empty) == false)
                {
                    yield return guid;
                }
            }
        }



        public bool RemovePersistedWorkflow(Guid instanceId)
        {
            string filename = GetFileName(instanceId);

            if (C1File.Exists(filename))
            {
                try
                {
                    C1File.Delete(filename);

                    LoggingService.LogVerbose("FileWorkFlowPersisetenceService", string.Format("Workflow persisted state deleted. Id = {0}", instanceId));
                }
                catch
                {
                    return false;
                }
            }

            return true;
        }



        protected override Activity LoadCompletedContextActivity(Guid scopeId, Activity outerActivity)
        {
            object obj = DeserializeActivity(outerActivity, scopeId);
            return (Activity)obj;
        }



        protected override Activity LoadWorkflowInstanceState(Guid instanceId)
        {
            string filename = GetFileName(instanceId);
            bool deleteFile = false;

            if (C1File.Exists(filename))
            {
                try
                {
                    object obj = DeserializeActivity(null, instanceId);
                    return (Activity)obj;
                }
                catch (Exception ex)
                {
                    LoggingService.LogCritical(LogTitle, ex);
                    deleteFile = true;
                }
            }

            if (deleteFile)
            {
                LoggingService.LogWarning(LogTitle, string.Format("Failed to load workflow with id '{0}'. Deleting file.", filename));
                C1File.Delete(filename);

                MarkWorkflowAsAborted(instanceId);                
            }

            return null;
        }



        protected override void SaveCompletedContextActivity(Activity activity)
        {
            Guid contextGuid = (Guid)activity.GetValue(Activity.ActivityContextGuidProperty);
            SerializeActivity(activity, contextGuid);
        }



        protected override void SaveWorkflowInstanceState(Activity rootActivity, bool unlock)
        {
            Guid contextGuid = (Guid)rootActivity.GetValue(Activity.ActivityContextGuidProperty);
            SerializeActivity(rootActivity, contextGuid);
        }


        protected override bool UnloadOnIdle(Activity activity)
        {
            return GetPersistingType(activity) == WorkflowPersistingType.Idle;
        }



        protected override void UnlockWorkflowInstanceState(Activity rootActivity)
        {
            //empty
        }

        private bool ActivityIsSerializable(Activity activity)
        {
            return PersistAll || GetPersistingType(activity) != WorkflowPersistingType.Never;
        }

        private WorkflowPersistingType GetPersistingType(Activity activity)
        {
            List<AllowPersistingWorkflowAttribute> attributes = activity.GetType().GetCustomAttributesRecursively<AllowPersistingWorkflowAttribute>().ToList();

            if (attributes.Count == 0) return WorkflowPersistingType.Never;

            return attributes[0].WorkflowPersistingType;
        }

        private void SerializeActivity(Activity rootActivity, Guid id)
        {
            if (!ActivityIsSerializable(rootActivity))
            {
                LoggingService.LogVerbose("FileWorkFlowPersisetenceService", string.Format("The workflow does not support persiting. Id = {0}, Type = {1}", id, rootActivity.GetType()));
                return;
            }

            string filename = GetFileName(id);

            IFormatter formatter = new BinaryFormatter();
            formatter.SurrogateSelector = ActivitySurrogateSelector.Default;

            using (C1FileStream stream = new C1FileStream(filename, FileMode.OpenOrCreate))
            {
                rootActivity.Save(stream, formatter);
                stream.Close();
            }

            // LoggingService.LogVerbose("FileWorkFlowPersisetenceService", string.Format("Workflow persisted. Id = {0}, Type = {1}", id, rootActivity.GetType()));
        }



        private object DeserializeActivity(Activity rootActivity, Guid id)
        {
            string filename = GetFileName(id);
            object result;

            IFormatter formatter = new BinaryFormatter();
            formatter.SurrogateSelector = ActivitySurrogateSelector.Default;

            using (C1FileStream stream = new C1FileStream(filename, FileMode.Open))
            {
                result = Activity.Load(stream, rootActivity, formatter);
                stream.Close();
            }

            // LoggingService.LogVerbose("FileWorkFlowPersisetenceService", string.Format("Workflow loaded. Id = {0}, Type = {1}", id, result.GetType()));

            return result;
        }

        private void MarkWorkflowAsAborted(Guid id)
        {
            lock (_syncRoot)
            {
                List<Guid> newList = new List<Guid>(_abortedWorkflows ?? new Guid[0]);
                newList.Add(id);

                _abortedWorkflows = newList.ToArray();
            }
        }

        private string GetFileName(Guid id)
        {
            return Path.Combine(this._baseDirectory, id.ToString() + ".bin");
        }
    }
}
