/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System.IO;
using Composite.Core.Implementation;


namespace Composite.Core.IO
{
    /// <summary>
    /// This class is a almost one to one version of System.IO.FileSystemWatcher. Using this implementation instead 
    /// of System.IO.FileSystemWatcher, will ensure that your code will work both on Standard Windows deployment 
    /// and Windows Azure deployment.
    /// See System.IO.FileSystemWatcher for more documentation on the methods of this class.
    /// See <see cref="Composite.Core.IO.Plugins.IOProvider.IC1FileStream"/>.
    /// </summary>
    public class C1FileSystemWatcher : ImplementationContainer<C1FileSystemWatcherImplementation>
    {
        /// <summary>
        /// Creates a new file system watcher given the path.
        /// </summary>
        /// <param name="path">Path to watch.</param>
        public C1FileSystemWatcher(string path)
            : this(path, null)
        {
        }



        /// <summary>
        /// Creates a new file system watcher given the path.
        /// </summary>
        /// <param name="path">Path to watch.</param>
        /// <param name="filter">Filter to use.</param>
        public C1FileSystemWatcher(string path, string filter)
            : base(() => ImplementationFactory.CurrentFactory.CreateC1FileSystemWatcher(path, filter))
        {
        }



        /// <summary>
        /// Gets or sets if events should be raised or not.
        /// </summary>
        public bool EnableRaisingEvents
        {
            get
            {
                return this.Implementation.EnableRaisingEvents;
            }
            set
            {
                this.Implementation.EnableRaisingEvents = value;
            }
        }



        /// <summary>
        /// Path to watch.
        /// </summary>
        public string Path
        {
            get
            {
                return this.Implementation.Path;
            }
            set
            {
                this.Implementation.Path = value;
            }
        }



        /// <summary>
        /// Filter to use.
        /// </summary>
        public string Filter
        {
            get
            {
                return this.Implementation.Filter;
            }
            set
            {
                this.Implementation.Filter = value;
            }
        }



        /// <summary>
        /// Gets or sets of subdirectories should also be watched.
        /// </summary>
        public bool IncludeSubdirectories
        {
            get
            {
                return this.Implementation.IncludeSubdirectories;
            }
            set
            {
                this.Implementation.IncludeSubdirectories = value;
            }
        }



        /// <summary>
        /// Adds or removes an event handler when new items are created.
        /// </summary>
        public event FileSystemEventHandler Created
        {
            add
            {
                this.Implementation.Created += value;
            }
            remove
            {
                this.Implementation.Created -= value;
            }
        }



        /// <summary>
        /// Adds or removes an event handler when new items changed.
        /// </summary>
        public event FileSystemEventHandler Changed
        {
            add
            {
                this.Implementation.Changed += value;
            }
            remove
            {
                this.Implementation.Changed -= value;
            }
        }



        /// <summary>
        /// Adds or removes an event handler when new items are renamed.
        /// </summary>
        public event RenamedEventHandler Renamed
        {
            add
            {
                this.Implementation.Renamed += value;
            }
            remove
            {
                this.Implementation.Renamed -= value;
            }
        }



        /// <summary>
        /// Adds or removes an event handler when new items are deleted.
        /// </summary>
        public event FileSystemEventHandler Deleted
        {
            add
            {
                this.Implementation.Deleted += value;
            }
            remove
            {
                this.Implementation.Deleted -= value;
            }
        }



        /// <summary>
        /// Adds or removes an event handler when an error occure.
        /// </summary>
        public event ErrorEventHandler Error
        {
            add
            {
                this.Implementation.Error += value;
            }
            remove
            {
                this.Implementation.Error -= value;
            }
        }



        /// <summary>
        /// Gets or sets the notify filter.
        /// </summary>
        public NotifyFilters NotifyFilter
        {
            get
            {
                return this.Implementation.NotifyFilter;
            }
            set
            {
                this.Implementation.NotifyFilter = value;
            }
        }



        /// <summary>
        /// Begins the initialization.
        /// </summary>
        public void BeginInit()
        {
            Log.LogInformation("C1FileSystemWatcher", "BeginInit starting");
            this.Implementation.BeginInit();
            Log.LogInformation("C1FileSystemWatcher", "BeginInit done");
        }



        /// <summary>
        /// Ends the initialization.
        /// </summary>
        public void EndInit()
        {
            this.Implementation.EndInit();
        }



        /// <summary>
        /// </summary>
        /// <param name="changeType"></param>
        /// <returns></returns>
        public C1WaitForChangedResult WaitForChanged(WatcherChangeTypes changeType)
        {
            return this.Implementation.WaitForChanged(changeType);
        }



        /// <summary>
        /// </summary>
        /// <param name="changeType"></param>
        /// <param name="timeout"></param>
        /// <returns></returns>
        public C1WaitForChangedResult WaitForChanged(WatcherChangeTypes changeType, int timeout)
        {
            return this.Implementation.WaitForChanged(changeType, timeout);
        }
    }
}
