/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using Composite.Core.Configuration;
using Composite.Core.IO;
using Composite.Core.Xml;


namespace Composite.Core.PackageSystem
{
    internal static class LicenseDefinitionManager
    {
        private const string LicenseFileExtension = ".license";
        private static readonly string LogTitle = typeof (LicenseDefinitionManager).Name;

        private static readonly string _packageLicenseDirectory;
        private static readonly int _maximumProductNameLength;

        static LicenseDefinitionManager()
        {
            _packageLicenseDirectory = PathUtil.Resolve(GlobalSettingsFacade.PackageLicenseDirectory);

            if (!C1Directory.Exists(_packageLicenseDirectory))
            {
                C1Directory.CreateDirectory(_packageLicenseDirectory);
            }

            _maximumProductNameLength = 255 - (GlobalSettingsFacade.MaximumRootPathLength + (GlobalSettingsFacade.PackageLicenseDirectory.Length - 1) + LicenseFileExtension.Length);
        }


        
        public static PackageLicenseDefinition GetLicenseDefinition(Guid productId)
        {
            return GetLicenseDefinitions(productId).OrderByDescending(l => l.Expires).FirstOrDefault();
        }


        public static PackageLicenseDefinition[] GetLicenseDefinitions(Guid productId)
        {
            var result = new List<PackageLicenseDefinition>();

            foreach (var file in C1Directory.GetFiles(_packageLicenseDirectory, "*" + LicenseFileExtension, SearchOption.TopDirectoryOnly))
            {
                var license = TryLoadLicenseFile(file);

                if (license != null && license.ProductId == productId)
                {
                    result.Add(license);
                }
            }

            string obsoloteFilename = GetObsoleteLicenseFilename(productId);
            if (C1File.Exists(obsoloteFilename))
            {
                var license = TryLoadLicenseFile(obsoloteFilename);

                if (license != null)
                {
                    if (license.ProductId == productId)
                    {
                        result.Add(license);
                    }
                    else
                    {
                        Log.LogError(LogTitle, "The license for the product '{0}' does not match the product in the license file '{1}'", productId, license.ProductId);
                    }
                }
            }

            return result.ToArray();
        }

        
        public static void StoreLicenseDefinition(PackageLicenseDefinition licenseDefinition)
        {
            XDocument doc = new XDocument(
                new XElement("License",
                    new XElement("Name", licenseDefinition.ProductName),
                    new XElement("InstallationId", licenseDefinition.InstallationId),
                    new XElement("ProductId", licenseDefinition.ProductId),
                    new XElement("Permanent", licenseDefinition.Permanent),                                        
                    new XElement("Expires", licenseDefinition.Expires),
                    new XElement("LicenseKey", licenseDefinition.LicenseKey),
                    new XElement("PurchaseUrl", licenseDefinition.PurchaseUrl)
                )
            );

            string filename = GetLicenseFilename(licenseDefinition);

            licenseDefinition.LicenseFileName = filename;

            doc.SaveToFile(filename);
        }


        private static PackageLicenseDefinition TryLoadLicenseFile(string filePath)
        {
            XDocument doc = XDocumentUtils.Load(filePath);

            var licenseDefinition = new PackageLicenseDefinition
            {
                ProductName = doc.Descendants("Name").Single().Value,
                InstallationId = (Guid)doc.Descendants("InstallationId").Single(),
                ProductId = (Guid)doc.Descendants("ProductId").Single(),
                Permanent = (bool)doc.Descendants("Permanent").Single(),
                Expires = (doc.Descendants("Expires").Any() ? (DateTime)doc.Descendants("Expires").Single() : DateTime.MaxValue),
                LicenseKey = doc.Descendants("LicenseKey").Single().Value,
                PurchaseUrl = (doc.Descendants("PurchaseUrl").Any() ? doc.Descendants("PurchaseUrl").Single().Value : ""),
                LicenseFileName = filePath
            };

            if (licenseDefinition.InstallationId != InstallationInformationFacade.InstallationId)
            {
                Log.LogError(LogTitle, string.Format("The license for the product '{0}' ({1}) does not match the current installation", licenseDefinition.ProductId, licenseDefinition.ProductName));
                return null;
            }

            return licenseDefinition;
        }

        public static void RemoveLicenseDefintion(Guid productId)
        {
            foreach (var license in GetLicenseDefinitions(productId))
            {
                FileUtils.Delete(license.LicenseFileName);
            }
        }


        private static string GetLicenseFilename(PackageLicenseDefinition packageLicenseDefinition)
        {
            string productName = packageLicenseDefinition.ProductName;

            if (productName.Length > _maximumProductNameLength)
            {
                productName = productName.Substring(productName.Length - _maximumProductNameLength);
            }

            return Path.Combine(_packageLicenseDirectory, productName + LicenseFileExtension);
        }


        private static string GetObsoleteLicenseFilename(Guid productId)
        {
            return Path.Combine(_packageLicenseDirectory, string.Format("{0}.xml", productId));
        }
    }
}
