/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.ServiceModel;
using Composite.Core.Logging;
using Composite.Core.PackageSystem.Foundation;
using Composite.Core.PackageSystem.WebServiceClient;


namespace Composite.Core.PackageSystem
{
    internal sealed class PackageServerFacadeImpl : IPackageServerFacade
    {
        private PackageServerFacadeImplCache _packageServerFacadeImplCache = new PackageServerFacadeImplCache();


        public ServerUrlValidationResult ValidateServerUrl(string packageServerUrl)
        {
            try
            {
                BasicHttpBinding basicHttpBinding = new BasicHttpBinding();
                basicHttpBinding.Security.Mode = BasicHttpSecurityMode.Transport;
                basicHttpBinding.MaxReceivedMessageSize = int.MaxValue;
                PackagesSoapClient client = new PackagesSoapClient(basicHttpBinding, new EndpointAddress(string.Format("https://{0}", packageServerUrl)));

                client.IsOperational();
                return ServerUrlValidationResult.Https;
            }
            catch (Exception)
            {
            }

            try
            {
                BasicHttpBinding basicHttpBinding = new BasicHttpBinding();
                basicHttpBinding.MaxReceivedMessageSize = int.MaxValue;
                PackagesSoapClient client = new PackagesSoapClient(new BasicHttpBinding(), new EndpointAddress(string.Format("http://{0}", packageServerUrl)));

                client.IsOperational();
                return ServerUrlValidationResult.Http;
            }
            catch (Exception)
            {
            }

            return ServerUrlValidationResult.Invalid;
        }



        public IEnumerable<PackageDescription> GetPackageDescriptions(string packageServerUrl, Guid installationId, CultureInfo userCulture)
        {
            List<PackageDescription> packageDescriptions = _packageServerFacadeImplCache.GetCachedPackageDescription(packageServerUrl, installationId, userCulture);
            if (packageDescriptions != null) return packageDescriptions;

            PackageDescriptor[] packageDescriptors = null;
            try
            {
                PackagesSoapClient client = CreateClient(packageServerUrl);

                packageDescriptors = client.GetPackageList(installationId, userCulture.ToString());
            }
            catch (Exception ex)
            {
                LoggingService.LogError("PackageServerFacade", ex);
            }

            packageDescriptions = new List<PackageDescription>();
            if (packageDescriptors != null)
            {
                foreach (PackageDescriptor packageDescriptor in packageDescriptors)
                {
                    if (ValidatePackageDescriptor(packageDescriptor))
                    {
                        packageDescriptions.Add(new PackageDescription
                        {
                            PackageFileDownloadUrl = packageDescriptor.PackageFileDownloadUrl,
                            PackageVersion = packageDescriptor.PackageVersion,
                            Description = packageDescriptor.Description,
                            EulaId = packageDescriptor.EulaId,
                            GroupName = packageDescriptor.GroupName,
                            Id = packageDescriptor.Id,
                            InstallationRequireLicenseFileUpdate = packageDescriptor.InstallationRequireLicenseFileUpdate,
                            IsFree = packageDescriptor.IsFree,
                            IsTrial = packageDescriptor.IsTrial,
                            LicenseRuleId = packageDescriptor.LicenseId,
                            MaxCompositeVersionSupported = packageDescriptor.MaxCompositeVersionSupported,
                            MinCompositeVersionSupported = packageDescriptor.MinCompositeVersionSupported,
                            Name = packageDescriptor.Name,
                            PriceAmmount = packageDescriptor.PriceAmmount,
                            PriceCurrency = packageDescriptor.PriceCurrency,
                            ReadMoreUrl = packageDescriptor.ReadMoreUrl,
                            TechicalDetails = packageDescriptor.TechicalDetails,
                            TrialPeriodDays = packageDescriptor.TrialPeriodDays,
                            UpgradeAgreementMandatory = packageDescriptor.UpgradeAgreementMandatory,
                            Vendor = packageDescriptor.Author
                        });
                    }
                }

                _packageServerFacadeImplCache.AddCachedPackageDescription(packageServerUrl, installationId, userCulture, packageDescriptions);
            }            

            return packageDescriptions;
        }



        public string GetEulaText(string packageServerUrl, Guid eulaId, CultureInfo userCulture)
        {
            PackagesSoapClient client = CreateClient(packageServerUrl);

            string eulaText = client.GetEulaText(eulaId, userCulture.ToString());

            return eulaText;
        }



        public Stream GetInstallFileStream(string packageFileDownloadUrl)
        {
            LoggingService.LogVerbose("PackageServerFacade", string.Format("Downloading file: {0}", packageFileDownloadUrl));

            System.Net.WebClient client = new System.Net.WebClient();
            return client.OpenRead(packageFileDownloadUrl);
        }



        public void RegisterPackageInstallationCompletion(string packageServerUrl, Guid installationId, Guid packageId, string localUserName, string localUserIp)
        {
            PackagesSoapClient client = CreateClient(packageServerUrl);

            client.RegisterPackageInstallationCompletion(installationId, packageId, localUserName, localUserIp);
        }



        public void RegisterPackageInstallationFailure(string packageServerUrl, Guid installationId, Guid packageId, string localUserName, string localUserIp, string exceptionString)
        {
            PackagesSoapClient client = CreateClient(packageServerUrl);

            client.RegisterPackageInstallationFailure(installationId, packageId, localUserName, localUserIp, exceptionString);
        }



        public void RegisterPackageUninstall(string packageServerUrl, Guid installationId, Guid packageId, string localUserName, string localUserIp)
        {
            PackagesSoapClient client = CreateClient(packageServerUrl);

            client.RegisterPackageUninstall(installationId, packageId, localUserName, localUserIp);
        }



        public void ClearCache()
        {
            _packageServerFacadeImplCache.Clear();
        }



        private bool ValidatePackageDescriptor(PackageDescriptor packageDescriptor)
        {
            string newVersion;
            if (VersionStringHelper.ValidateVersion(packageDescriptor.PackageVersion, out newVersion) == false)
            {
                LoggingService.LogWarning("PackageServerFacade", string.Format("The package '{0}' ({1}) did not validate and is skipped", packageDescriptor.Name, packageDescriptor.Id));
                return false;
            }
            else
            {
                packageDescriptor.PackageVersion = newVersion;
            }

            if (VersionStringHelper.ValidateVersion(packageDescriptor.MinCompositeVersionSupported, out newVersion) == false)
            {
                LoggingService.LogWarning("PackageServerFacade", string.Format("The package '{0}' ({1}) did not validate and is skipped", packageDescriptor.Name, packageDescriptor.Id));
                return false;
            }
            else
            {
                packageDescriptor.MinCompositeVersionSupported = newVersion;
            }

            if (VersionStringHelper.ValidateVersion(packageDescriptor.MaxCompositeVersionSupported, out newVersion) == false)
            {
                LoggingService.LogWarning("PackageServerFacade", string.Format("The package '{0}' ({1}) did not validate and is skipped", packageDescriptor.Name, packageDescriptor.Id));
                return false;
            }
            else
            {
                packageDescriptor.MaxCompositeVersionSupported = newVersion;
            }

            return true;
        }
      


        private PackagesSoapClient CreateClient(string packageServerUrl)
        {
            BasicHttpBinding basicHttpBinding = new BasicHttpBinding();

            if (RuntimeInformation.IsDebugBuild)
            {
                basicHttpBinding.CloseTimeout = TimeSpan.FromSeconds(1);
                basicHttpBinding.OpenTimeout = TimeSpan.FromSeconds(1);
                basicHttpBinding.ReceiveTimeout = TimeSpan.FromSeconds(1);
                basicHttpBinding.SendTimeout = TimeSpan.FromSeconds(1);

                basicHttpBinding.CloseTimeout = TimeSpan.FromMinutes(2);
                basicHttpBinding.OpenTimeout = TimeSpan.FromMinutes(2);
                basicHttpBinding.ReceiveTimeout = TimeSpan.FromMinutes(2);
                basicHttpBinding.SendTimeout = TimeSpan.FromMinutes(2);
            }
            else
            {
                basicHttpBinding.CloseTimeout = TimeSpan.FromMinutes(1);
                basicHttpBinding.OpenTimeout = TimeSpan.FromMinutes(1);
                basicHttpBinding.ReceiveTimeout = TimeSpan.FromMinutes(1);
                basicHttpBinding.SendTimeout = TimeSpan.FromMinutes(1);
            }

            if (packageServerUrl.StartsWith("https"))
            {
                basicHttpBinding.Security.Mode = BasicHttpSecurityMode.Transport;
            }

            basicHttpBinding.MaxReceivedMessageSize = int.MaxValue;

            PackagesSoapClient client = new PackagesSoapClient(basicHttpBinding, new EndpointAddress(packageServerUrl));

            return client;
        }
    }
}
