/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using Composite.Data;
using Composite.Core.IO.Zip;


namespace Composite.Core.PackageSystem
{
    /// <summary>    
    /// </summary>
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)] 
    public sealed class PackageUninstallerContext
    {
        private Dictionary<Type, Dictionary<DataScopeIdentifier, Dictionary<CultureInfo, List<DataKeyPropertyCollection>>>> _dataPendingForDeletion = new Dictionary<Type, Dictionary<DataScopeIdentifier, Dictionary<CultureInfo, List<DataKeyPropertyCollection>>>>();


        internal PackageUninstallerContext(IZipFileSystem zipFileSystem, PackageInformation packageInformation)
        {
            if (zipFileSystem == null) throw new ArgumentNullException("zipFileSystem");

            this.ZipFileSystem = zipFileSystem;
            this.PackageInformation = packageInformation;
        }


        /// <exclude />
        public IZipFileSystem ZipFileSystem { get; private set; }

        /// <exclude />
        public PackageInformation PackageInformation { get; private set; }


        /// <exclude />
        public void AddPendingForDeletionData(Type interfaceType, DataScopeIdentifier dataScopeIdentifier, CultureInfo locale, DataKeyPropertyCollection dataKeyPropertyCollection)
        {
            if (interfaceType == null) throw new ArgumentNullException("interfaceType");
            if (dataScopeIdentifier == null) throw new ArgumentNullException("dataScopeIdentifier");
            if (dataKeyPropertyCollection == null) throw new ArgumentNullException("dataKeyPropertyCollection");

            List<DataKeyPropertyCollection> dataKeyPropertyCollections = GetDataKeyPropertyCollections(interfaceType, dataScopeIdentifier, locale);

            if (dataKeyPropertyCollections.Contains(dataKeyPropertyCollection))
            {
                throw new ArgumentException(string.Format("The data item of type '{0}' with the key '{1}' has already been added", interfaceType, dataKeyPropertyCollection));
            }

            dataKeyPropertyCollections.Add(dataKeyPropertyCollection);
        }



        /// <exclude />
        public bool IsPendingForDeletionData(IData data)
        {
            if (data == null) throw new ArgumentNullException("data");

            DataKeyPropertyCollection dataKeyPropertyCollection = data.CreateDataKeyPropertyCollection();

            return IsPendingForDeletionData(data.DataSourceId.InterfaceType, data.DataSourceId.DataScopeIdentifier, data.DataSourceId.LocaleScope, dataKeyPropertyCollection);
        }



        /// <exclude />
        public bool IsPendingForDeletionData(Type interfaceType, DataScopeIdentifier dataScopeIdentifier, CultureInfo locale, DataKeyPropertyCollection dataKeyPropertyCollection)
        {
            if (interfaceType == null) throw new ArgumentNullException("interfaceType");
            if (dataScopeIdentifier == null) throw new ArgumentNullException("dataScopeIdentifier");
            if (dataKeyPropertyCollection == null) throw new ArgumentNullException("dataKeyPropertyCollection");

            List<DataKeyPropertyCollection> dataKeyPropertyCollections = GetDataKeyPropertyCollections(interfaceType, dataScopeIdentifier, locale);

            return dataKeyPropertyCollections.Contains(dataKeyPropertyCollection);
        }



        private List<DataKeyPropertyCollection> GetDataKeyPropertyCollections(Type interfaceType, DataScopeIdentifier dataScopeIdentifier, CultureInfo locale)
        {
            Dictionary<DataScopeIdentifier, Dictionary<CultureInfo, List<DataKeyPropertyCollection>>> dataKeyPropertyCollectionDataScopes;
            if (_dataPendingForDeletion.TryGetValue(interfaceType, out dataKeyPropertyCollectionDataScopes) == false)
            {
                dataKeyPropertyCollectionDataScopes = new Dictionary<DataScopeIdentifier, Dictionary<CultureInfo, List<DataKeyPropertyCollection>>>();
                _dataPendingForDeletion.Add(interfaceType, dataKeyPropertyCollectionDataScopes);
            }

            Dictionary<CultureInfo, List<DataKeyPropertyCollection>> dataKeyPropertyCollectionsLoclaes;
            if (dataKeyPropertyCollectionDataScopes.TryGetValue(dataScopeIdentifier, out dataKeyPropertyCollectionsLoclaes) == false)
            {
                dataKeyPropertyCollectionsLoclaes = new Dictionary<CultureInfo, List<DataKeyPropertyCollection>>();
                dataKeyPropertyCollectionDataScopes.Add(dataScopeIdentifier, dataKeyPropertyCollectionsLoclaes);
            }

            if (locale == null)
            {
                locale = CultureInfo.InvariantCulture;
            }

            List<DataKeyPropertyCollection> dataKeyPropertyCollections;
            if (dataKeyPropertyCollectionsLoclaes.TryGetValue(locale, out dataKeyPropertyCollections) == false)
            {
                dataKeyPropertyCollections = new List<DataKeyPropertyCollection>();
                dataKeyPropertyCollectionsLoclaes.Add(locale, dataKeyPropertyCollections);
            }

            return dataKeyPropertyCollections;
        }
    }
}
