/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using Composite.Core.Configuration;
using Composite.Core.Extensions;
using Composite.Core.ResourceSystem.Foundation.PluginFacades;
using Composite.Core.ResourceSystem.Plugins.ResourceProvider;
using Composite.Core.ResourceSystem.Plugins.ResourceProvider.Runtime;
using Composite.Core.Collections.Generic;
using Composite.C1Console.Events;
using Composite.Plugins.ResourceSystem.AggregationLocalizationProvider;

namespace Composite.Core.ResourceSystem.Foundation
{
    internal static class ResourceProviderRegistry
    {
        private static readonly ResourceLocker<Resources> _resourceLocker = new ResourceLocker<Resources>(new Resources(), Resources.Initialize, false);
        private static readonly string LogTitle = typeof (ResourceProviderRegistry).Name;


        static ResourceProviderRegistry()
        {
            GlobalEventSystemFacade.SubscribeToFlushEvent((a) => Flush());
        }



        public static IEnumerable<string> IconResourceProviderNames
        {
            get
            {
                using (_resourceLocker.ReadLocker)
                {
                    return _resourceLocker.Resources.IconResourceProviders;
                }
            }
        }



        public static IEnumerable<string> StringResourceProviderNames
        {
            get
            {
                using (_resourceLocker.ReadLocker)
                {
                    return _resourceLocker.Resources.StringResourceProviders;
                }
            }
        }


        public static IEnumerable<string> LocalizationProviderNames
        {
            get
            {
                using (_resourceLocker.ReadLocker)
                {
                    return _resourceLocker.Resources.LocalizationProviders;
                }
            }
        }



        private static void Flush()
        {
            _resourceLocker.ResetInitialization();
        }


        private sealed class Resources
        {
            public List<string> IconResourceProviders;
            public List<string> StringResourceProviders;
            public List<string> LocalizationProviders;

            public static void Initialize(Resources resources)
            {
                resources.IconResourceProviders = new List<string>();
                resources.StringResourceProviders = new List<string>();
                resources.LocalizationProviders = new List<string>();

                if ((ConfigurationServices.ConfigurationSource != null) &&
                    (ConfigurationServices.ConfigurationSource.GetSection(ResourceProviderSettings.SectionName) != null))
                {
                    ResourceProviderSettings configuration = (ResourceProviderSettings)ConfigurationServices.ConfigurationSource.GetSection(ResourceProviderSettings.SectionName);

                    foreach (ResourceProviderData data in configuration.ResourceProviderPlugins)
                    {
                        Type type = ResourceProviderPluginFacade.GetProviderType(data.Name);

                        if (typeof(IIconResourceProvider).IsAssignableFrom(type))
                        {
                            resources.IconResourceProviders.Add(data.Name);
                        }
                        else if (typeof(IStringResourceProvider).IsAssignableFrom(type))
                        {
                            Log.LogWarning(LogTitle, ("String resource provider '{0}' definition ignored." +
                                                      "\nEither remove it from Composite.config, or move the provider definition under a provider of type '{1}' ")
                                                      .FormatWith(data.Name, typeof(AggregationLocalizationProvider).FullName));

                            resources.StringResourceProviders.Add(data.Name);
                        }
                        else if (typeof(ILocalizationProvider).IsAssignableFrom(type))
                        {
                            resources.LocalizationProviders.Add(data.Name);
                        }
                        else 
                        {
                            throw new NotSupportedException(string.Format("Unkown resource provider type '{0}'", type));
                        }
                    }
                }
            }
        }
    }
}
