/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Reflection;
using Composite.Core.Collections.Generic;
using Composite.Data;
using System.Xml.Linq;
using Composite.Data.Types;
using Composite.Core.Logging;
using Composite.Core.Extensions;
using Composite.Core.Xml;
using Composite.Core.WebClient;
using System.Web;

namespace Composite.Core.WebClient.Renderings.Data
{
    /// <summary>    
    /// </summary>
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)]
    public class KeyTemplatedXhtmlRendererAttribute : XhtmlRendererProviderAttribute
    {
        private XhtmlRenderingType _supportedRenderingType;
        private IDataXhtmlRenderer _renderer;
        private XhtmlRenderingEncoding _renderingEncoding;

        /// <summary>
        /// Created a XHTML Renderer that uses the specified template to create markup.
        /// The key of the data will be inserted into the specified template where '{id}' is.
        /// If you spcify '{label}' the system will fetch the value of the label field and insert it.
        /// You also can use '{field:__a field name__}' syntax to insert a field value.
        /// Use '~' to create absolute paths.
        /// Example: <example>&lt;a href='~/showProduct.aspx?id={id}'>read more about {label}&lt;/a></example>
        /// </summary>
        public KeyTemplatedXhtmlRendererAttribute(XhtmlRenderingType renderingType, string formatedTemplate)
        {
            this.FormatedTemplate = formatedTemplate;
            _supportedRenderingType = renderingType;
            _renderingEncoding = XhtmlRenderingEncoding.None;
        }


        /// <summary>
        /// Created a XHTML Renderer that uses the specified template to create markup.
        /// The key of the data will be inserted into the specified template where '{id}' is.
        /// If you spcify '{label}' the system will fetch the value of the label field and insert it.
        /// You also can use '{field:__a field name__}' syntax to insert a field value.
        /// Use '~' to create absolute paths.
        /// Example: <example>&lt;a href='~/showProduct.aspx?id={id}'>read more about {label}&lt;/a></example>
        /// </summary>
        public KeyTemplatedXhtmlRendererAttribute(XhtmlRenderingType renderingType, XhtmlRenderingEncoding renderingEncoding, string formatedTemplate)
        {
            this.FormatedTemplate = formatedTemplate;
            _supportedRenderingType = renderingType;
            _renderingEncoding = renderingEncoding;
        }


        /// <exclude />
        public string FormatedTemplate
        {
            get;
            private set;
        }


        /// <exclude />
        public override IDataXhtmlRenderer BuildRenderer()
        {
            if (_renderer == null)
            {
                _renderer = new KeyBasedXhtmlRenderer(this.FormatedTemplate, _renderingEncoding);
            }
            return _renderer;
        }


        /// <exclude />
        public override XhtmlRenderingType SupportedRenderingType
        {
            get
            {
                return _supportedRenderingType;
            }
        }


        private class KeyBasedXhtmlRenderer : IDataXhtmlRenderer
        {
            private static readonly string LogTitle = "KeyBasedXhtmlRenderer";

            private readonly string _templateString;

            private readonly List<string> _fieldNames = new List<string>();
            private readonly List<string> _fieldPatterns = new List<string>();
            private readonly bool _labelHasToBeEvaluated;
            private readonly bool _idHasToBeEvaluated;
            private readonly XhtmlRenderingEncoding _renderingEncoding;

            private readonly Hashtable<Type, Hashtable<string, PropertyInfo>> _reflectionCache = new Hashtable<Type, Hashtable<string, PropertyInfo>>();

            public KeyBasedXhtmlRenderer(string templateString, XhtmlRenderingEncoding renderingEncoding)
            {
                templateString = templateString.Replace("~", UrlUtils.PublicRootPath);
                templateString = templateString.Replace("{label}", "{0}");
                templateString = templateString.Replace("{id}", "{1}");

                _labelHasToBeEvaluated = templateString.Contains("{0}");
                _idHasToBeEvaluated = templateString.Contains("{1}");
                _renderingEncoding = renderingEncoding;

                int stringFormattingIndex = 2;

                while (true)
                {
                    int fieldDefinitionOffset = templateString.IndexOf("{field:");
                    if (fieldDefinitionOffset < 0) break;

                    int closingBraceIndex = templateString.IndexOf("}", fieldDefinitionOffset + 7);
                    Verify.That(closingBraceIndex > 0, "Invalid rendering template.");

                    string fieldName = templateString.Substring(fieldDefinitionOffset + 7, closingBraceIndex - fieldDefinitionOffset - 7);
                    _fieldNames.Add(fieldName);

                    string fieldPattern = templateString.Substring(fieldDefinitionOffset, closingBraceIndex - fieldDefinitionOffset + 1);
                    _fieldPatterns.Add(fieldPattern);

                    templateString = templateString.Replace(fieldPattern, "{" + (stringFormattingIndex++) + "}");
                }

                _templateString = string.Format("<body xmlns='{0}'>{1}</body>", Namespaces.Xhtml, templateString);
            }


            public XhtmlDocument Render(IDataReference dataReferenceToRender)
            {
                if (!dataReferenceToRender.IsSet)
                {
                    return new XhtmlDocument();
                }

                IData dataToRender;

                if (dataReferenceToRender.ReferencedType == typeof(IPage))
                {
                    dataToRender = DataFacade.TryGetDataByUniqueKey<IPage>(dataReferenceToRender.KeyValue);
                    if (dataToRender == null)
                    {
                        return new XhtmlDocument();
                    }
                }
                else
                {
                    dataToRender = dataReferenceToRender.Data;
                }

                string markup = _templateString;
                string labelEncoded = "";
                string keyValue = "";


                if (_labelHasToBeEvaluated)
                {
                    labelEncoded = HttpUtility.HtmlEncode(dataToRender.GetLabel());
                }

                if (_idHasToBeEvaluated)
                {
                    keyValue = dataReferenceToRender.KeyValue.ToString();
                }

                var parameters = new object[2 + _fieldNames.Count];

                parameters[0] = labelEncoded;
                parameters[1] = keyValue;

                // Getting field values
                for (int i = 0; i < _fieldNames.Count; i++)
                {
                    Type referenceType = dataReferenceToRender.ReferencedType;

                    var propertiesMap = _reflectionCache.EnsureValue(referenceType,
                        () => new Hashtable<string, PropertyInfo>());

                    string fieldName = _fieldNames[i];

                    PropertyInfo propertyInfo;
                    if (!propertiesMap.TryGetValue(fieldName, out propertyInfo))
                    {
                        lock (propertiesMap)
                        {
                            if (!propertiesMap.TryGetValue(fieldName, out propertyInfo))
                            {
                                Type type = referenceType;

                                while (type != null && type != typeof(object))
                                {
                                    propertyInfo = type.GetProperty(fieldName);
                                    if (propertyInfo != null)
                                    {
                                        propertiesMap.Add(fieldName, propertyInfo);
                                        break;
                                    }

                                    if (type.GetInterfaces().Length > 0)
                                    {
                                        type = type.GetInterfaces()[0];
                                    }
                                }

                                if (propertyInfo == null)
                                {
                                    LoggingService.LogWarning(LogTitle, "Failed to find property '{0}' on type '{1}'"
                                        .FormatWith(fieldName, referenceType.FullName));

                                    propertiesMap.Add(fieldName, null);
                                }
                            }
                        }
                    }

                    string value = String.Empty;

                    if (propertyInfo != null)
                    {
                        value = (propertyInfo.GetValue(dataToRender, new object[0]) ?? String.Empty).ToString();
                    }

                    switch (_renderingEncoding)
                    {
                        case XhtmlRenderingEncoding.None:
                            break;
                        case XhtmlRenderingEncoding.AttributeContent:
                            value = HttpUtility.HtmlAttributeEncode(value);
                            break;
                        case XhtmlRenderingEncoding.TextContent:
                            value = HttpUtility.HtmlEncode(value);
                            break;
                        default:
                            throw new NotImplementedException("Unexpected XhtmlRenderingEncoding value");
                    }

                    parameters[2 + i] = value;
                }

                string evaluatedMarkup = string.Format(markup, parameters);

                XElement bodyMarkup = XElement.Parse(evaluatedMarkup);

                XhtmlDocument xhtmlDocument = new XhtmlDocument();

                xhtmlDocument.Body.Add(bodyMarkup.Nodes());

                return xhtmlDocument;
            }
        }
    }
}
