/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Threading;
using Composite.Core.Collections.Generic;
using Composite.Core.Types;
using Composite.Data.Caching.Foundation;
using System.Reflection;
using Composite.Core.Linq;
using Composite.Data.Foundation;


namespace Composite.Data.Caching
{
    internal sealed class CachingEnumerable<T> : IEnumerable<T>
    {
        private readonly IEnumerable<T> _enumerable;

        public CachingEnumerable(IEnumerable<T> enumerable)
        {
            _enumerable = enumerable;
        }


        public IEnumerator<T> GetEnumerator()
        {
            return new CachingEnumerator<T>(_enumerable.GetEnumerator());
        }


        IEnumerator IEnumerable.GetEnumerator()
        {
            return new CachingEnumerator<T>(_enumerable.GetEnumerator());
        }
    }

    internal interface ICachedQuery
    {
        IQueryable GetOriginalQuery();
    }

    /// <summary>
    /// Used for optimizing execution of GetDataByUniqueKey method
    /// </summary>
    internal interface CachingQueryable_CachedByKey
    {
        IData GetCachedValueByKey(object key);
    }


    internal sealed class CachingQueryable<T> : ICachedQuery, IOrderedQueryable<T>, IQueryProvider, ICachingQueryable, CachingQueryable_CachedByKey
    {
        private readonly IQueryable _source;
        private readonly Expression _currentExpression;
        private readonly CachingEnumerable<T> _wrappedEnumerable;
        private readonly Func<IQueryable> _getQueryFunc;

        private DataCachingFacade.CachedTable _cachedTable;
        private static readonly MethodInfo _wrappingMethodInfo;

        private IEnumerable<T> _innerEnumerable;

        static CachingQueryable()
        {
            if (typeof(IData).IsAssignableFrom(typeof(T)))
            {
                _wrappingMethodInfo = StaticReflection.GetGenericMethodInfo(a => DataWrappingFacade.Wrap((IData) a))
                                                      .MakeGenericMethod(new[] {typeof (T)});
            }
        } 

        public CachingQueryable(DataCachingFacade.CachedTable cachedTable, Func<IQueryable> originalQueryGetter)
        {
            _cachedTable = cachedTable;

            _source = cachedTable.Queryable;
            _currentExpression = Expression.Constant(this);
            _getQueryFunc = originalQueryGetter;

            if (_source is IEnumerable<T>)
            {
                _innerEnumerable = _source as IEnumerable<T>;
                if(_innerEnumerable is CachingEnumerable<T>)
                {
                    _wrappedEnumerable = _innerEnumerable as CachingEnumerable<T>;
                }
            }
        }


        public CachingQueryable(IQueryable source, Expression currentExpression, Func<IQueryable> originalQueryGetter)
        {
            _source = source;
            _currentExpression = currentExpression;
            _getQueryFunc = originalQueryGetter;
        }



        #region Generic methods

        public IQueryable<S> CreateQuery<S>(Expression expression)
        {
            return new CachingQueryable<S>(_source, expression, _getQueryFunc);
        }

        public IEnumerator<T> GetEnumerator()
        {
            var enumerable = BuildEnumerable();

            if (_wrappedEnumerable != null)
            {
                return _wrappedEnumerable.GetEnumerator();
            }

            var enumerator = enumerable.GetEnumerator();

            if (_source.ElementType == typeof(T))
            {
                return new CachingEnumerator<T>(enumerator);
            }
            return enumerator;
        }


        private IEnumerable<T> BuildEnumerable()
        {
            if (_innerEnumerable == null) {
                lock (this) {
                    if (_innerEnumerable == null) {
                        var visitor = new ChangeSourceExpressionVisitor();

                        Expression newExpression = visitor.Visit(_currentExpression);

                        var result = (IQueryable<T>)_source.Provider.CreateQuery(newExpression);

                        Verify.IsNotNull(result, "Failed to create an enumerator");

                        Thread.MemoryBarrier();
                        _innerEnumerable = result;
                    }
                }
            }

            return _innerEnumerable;
        }

        public S Execute<S>(Expression expression)
        {
            ChangeSourceExpressionVisitor visitor;

            if (_source.ElementType == typeof(S))
            {
                IQueryable<S> newSouce = new CachingEnumerable<S>(_source.Cast<S>()).AsQueryable();

                visitor = new ChangeSourceExpressionVisitor(newSouce.Expression);

            }
            else
            {
                visitor = new ChangeSourceExpressionVisitor();
            }

            Expression newExpression = visitor.Visit(expression);

            return (S)_source.Provider.Execute(newExpression);
        }

        #endregion

        #region Non generic methods

        IEnumerator IEnumerable.GetEnumerator()
        {
            MethodInfo methodInfo = CachingQueryableCache.GetCachingQueryableGetEnumeratorMethodInfo(typeof(T));

            return (IEnumerator)methodInfo.Invoke(this, null);
        }



        public IQueryable CreateQuery(Expression expression)
        {
            if (_currentExpression == expression) return this;

            Type elementType = TypeHelpers.FindElementType(expression);

            Type queryableType = CachingQueryableCache.GetCachingQueryableType(elementType);

            return (IQueryable)Activator.CreateInstance(
                queryableType,
                new object[] { _source, expression, _getQueryFunc });
        }



        public object Execute(Expression expression)
        {
            MethodInfo methodInfo = CachingQueryableCache.GetCachingQueryableExecuteMethodInfo(typeof(T), expression.Type);

            return methodInfo.Invoke(this, new object[] { expression });
        }

        #endregion


        public IData GetCachedValueByKey(object key)
        {
            if (_cachedTable.RowByKey == null)
            {
                lock(_cachedTable)
                {
                    if (_cachedTable.RowByKey == null)
                    {
                        var table = new Hashtable<object, object>();

                        PropertyInfo keyPropertyInfo = DataAttributeFacade.GetKeyProperties(typeof(T)).Single();

                        IEnumerable<T> enumerable = BuildEnumerable();

                        var emptyIndexCollection = new object[0];

                        foreach(T row in enumerable)
                        {
                            object rowKey = keyPropertyInfo.GetValue(row, emptyIndexCollection);

                            table.Add(rowKey, row);
                        }

                        _cachedTable.RowByKey = table;
                    }
                }
            }

            object cachedRow = _cachedTable.RowByKey[key];
            if (cachedRow == null) return null;

            return _wrappingMethodInfo.Invoke(null, new object[] { cachedRow }) as IData;
        }


        public Expression Expression
        {
            get { return _currentExpression; }
        }



        public Type ElementType
        {
            get { return typeof(T); }
        }



        public IQueryProvider Provider
        {
            get { return this; }
        }

        public IQueryable Source
        {
            get { return _source; }
        }

        public IQueryable GetOriginalQuery()
        {
            return _getQueryFunc();
        }
    }
}
