/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Linq;
using Composite.Core.Types;
using Composite.Data.DynamicTypes;


namespace Composite.Data
{
    /// <summary>
    /// The attribute will tell the system that a data property is a reference (foreign key) to another IData.
    /// </summary>
    /// <example> This sample shows how to use the ForeignKey attribute. 
    /// <code>
    /// // data interface attributes ...
    /// interface IMyDataType : IData
    /// {
    ///     [StoreFieldType(PhysicalStoreFieldType.Guid)]
    ///     [ImmutableFieldId("{D75EA67F-AD14-4BAB-8547-6D87002809F2}")]
    ///     [ForeignKey(typeof(IPage), "Id", AllowCascadeDeletes = true)]
    ///     Guid PageId { get; set; }
    ///     
    ///     // more data properties ...
    ///     
    /// }
    /// </code>
    /// </example>
    [AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = true)]
    public sealed class ForeignKeyAttribute : Attribute
    {
        private string _interfaceTypeManagerName = null;

        private Type _interfaceType = null;
        private string _keyPropertyName = null;

        private object _nullReferenceValue;
        private bool _isNullReferenceValueSet = false;

        private object _lock = new object();



        /// <summary>
        /// Tell the system that this data property is a reference (foreign key) to another IData.
        /// </summary>
        /// <param name="interfaceType">The type being referenced</param>
        /// <param name="keyPropertyName">The field being referenced</param>
        public ForeignKeyAttribute(Type interfaceType, string keyPropertyName)
        {
            _interfaceType = interfaceType;
            _keyPropertyName = keyPropertyName;
        }


        /// <summary>
        /// Only use this constructor for types that are registred with the DynamicTypeManager. Primary key field is infered.
        /// </summary>
        /// <param name="interfaceTypeManagerName">A string that will yield a type from the TypeManager.</param>
        public ForeignKeyAttribute(string interfaceTypeManagerName)
        {
            if (string.IsNullOrEmpty(interfaceTypeManagerName)) throw new ArgumentNullException("interfaceTypeManagerName");

            _interfaceTypeManagerName = interfaceTypeManagerName;
        }



        /// <summary>
        /// If the "parent" data is deleted and this is set to true, then the datas that 
        /// references the parent is also deleted.
        /// </summary>
        public bool AllowCascadeDeletes { get; set; }


        /// <summary>
        /// This value is used when foreign key integrity is performed.
        /// If this is not set, the data that the foreign key is pointing to must always exists.        
        /// </summary>
        public object NullReferenceValue
        {
            get
            {
                return _nullReferenceValue;
            }
            set
            {
                _nullReferenceValue = value;
                _isNullReferenceValueSet = true;
            }
        }


        /// <summary>
        /// The NullReferenceValue will be converted to the the type specifed with the property
        /// </summary>
        public Type NullReferenceValueType
        {
            get;
            set;
        }



        /// <summary>
        /// Use this if non-reference is allowed with strings foreign keys
        /// </summary>
        public bool NullableString
        {
            get;
            set;
        }



        /// <exclude />
        public bool IsNullReferenceValueSet
        {
            get { return _isNullReferenceValueSet; }
        }



        /// <exclude />
        public Type InterfaceType
        {
            get
            {
                lock (_lock)
                {
                    if (_interfaceType == null)
                    {
                        _interfaceType = TypeManager.GetType(_interfaceTypeManagerName);
                    }
                }

                return _interfaceType;
            }
        }



        /// <exclude />
        public bool IsValid
        {
            get
            {
                lock (_lock)
                {
                    if (_interfaceType == null)
                    {
                        _interfaceType = TypeManager.TryGetType(_interfaceTypeManagerName);
                    }
                }

                return _interfaceType != null;
            }
        }



        /// <exclude />
        public string TypeManagerName
        {
            get
            {
                lock (_lock)
                {
                    if (string.IsNullOrEmpty(_interfaceTypeManagerName))
                    {
                        _interfaceTypeManagerName = TypeManager.TrySerializeType(_interfaceType);
                    }
                }

                return _interfaceTypeManagerName;
            }
        }



        /// <exclude />
        public string KeyPropertyName
        {
            get
            {
                lock (_lock)
                {
                    if (_keyPropertyName == null)
                    {
                        _keyPropertyName = DynamicTypeManager.GetDataTypeDescriptor(this.InterfaceType).KeyPropertyNames.Single();
                    }
                }

                return _keyPropertyName;
            }
        }
    }
}
