/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.CodeDom;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using Composite.Core.Types;
using Composite.Data.DynamicTypes;


namespace Composite.Data.Foundation.CodeGeneration
{
    /// <summary>
    /// This class genereated code for data wrapper classes.
    /// </summary>
    internal static class DataWrapperCodeGenerator
    {
        private const string NamespaceName = "CompositeGenerated.DataWrappers";
        private const string WrappedObjectName = "_wrappedData";



        internal static void AddDataWrapperClassCode(CodeGenerationBuilder codeGenerationBuilder, Type interfaceType)
        {
            codeGenerationBuilder.AddReference(interfaceType.Assembly);

            DataTypeDescriptor dataTypeDescriptor = DataMetaDataFacade.GetDataTypeDescriptor(interfaceType.GetImmutableTypeId());

            AddDataWrapperClassCode(codeGenerationBuilder, dataTypeDescriptor);
        }



        internal static void AddDataWrapperClassCode(CodeGenerationBuilder codeGenerationBuilder, DataTypeDescriptor dataTypeDescriptor)
        {
            Type interfaceType = DataTypeTypesManager.GetDataType(dataTypeDescriptor);
            if (interfaceType == null) return;

            codeGenerationBuilder.AddReference(typeof(IDataWrapper).Assembly);
            codeGenerationBuilder.AddReference(typeof(EditorBrowsableAttribute).Assembly);

            CodeTypeDeclaration codeTypeDeclaration = CreateCodeTypeDeclaration(dataTypeDescriptor);

            codeGenerationBuilder.AddType(NamespaceName, codeTypeDeclaration);
        }



        internal static string CreateWrapperClassFullName(string interfaceTypeFullName)
        {
            return NamespaceName + "." + CreateWrapperClassName(interfaceTypeFullName);
        }



        private static string CreateWrapperClassName(string interfaceTypeFullName)
        {
            return string.Format("{0}Wrapper", interfaceTypeFullName.Replace('.', '_').Replace('+', '_'));
        }



        private static CodeTypeDeclaration CreateCodeTypeDeclaration(DataTypeDescriptor dataTypeDescriptor)
        {
            string fullName = dataTypeDescriptor.GetFullInterfaceName();

            IEnumerable<Tuple<string, Type, bool>> properties =
                dataTypeDescriptor.Fields.
                Select(f => new Tuple<string, Type, bool>(f.Name, f.InstanceType, f.IsReadOnly)).
                Concat(new[] { new Tuple<string, Type, bool>("DataSourceId", typeof(DataSourceId), true) });

            return CreateCodeTypeDeclaration(fullName, properties);
        }



        /// <summary>
        /// 
        /// </summary>
        /// <param name="interfaceTypeFullName"></param>
        /// <param name="properties">Tuple(string propertyName, Type propertyType, bool readOnly)</param>
        /// <returns></returns>
        private static CodeTypeDeclaration CreateCodeTypeDeclaration(string interfaceTypeFullName, IEnumerable<Tuple<string, Type, bool>> properties)
        {
            CodeTypeDeclaration declaration = new CodeTypeDeclaration();
            declaration.Name = CreateWrapperClassName(interfaceTypeFullName);
            declaration.IsClass = true;
            declaration.TypeAttributes = TypeAttributes.Public | TypeAttributes.Sealed;
            declaration.BaseTypes.Add(interfaceTypeFullName);
            declaration.BaseTypes.Add(typeof(IDataWrapper));
            declaration.CustomAttributes.Add(
                new CodeAttributeDeclaration(
                    new CodeTypeReference(typeof(EditorBrowsableAttribute)),
                    new CodeAttributeArgument(
                        new CodeFieldReferenceExpression(
                            new CodeTypeReferenceExpression(typeof(EditorBrowsableState)),
                            EditorBrowsableState.Never.ToString()
                        )
                    )
                )
            );

            declaration.Members.Add(new CodeMemberField(interfaceTypeFullName, WrappedObjectName));

            AddConstructor(declaration, interfaceTypeFullName);
            AddInterfaceProperties(declaration, properties);

            AddMethods(declaration, properties);

            return declaration;
        }



        private static void AddConstructor(CodeTypeDeclaration declaration, string interfaceTypeFullName)
        {
            const string parameterName = "data";

            CodeConstructor codeConstructor = new CodeConstructor();
            codeConstructor.Attributes = MemberAttributes.Public;
            codeConstructor.Parameters.Add(new CodeParameterDeclarationExpression(interfaceTypeFullName, parameterName));

            codeConstructor.Statements.Add(
                new CodeAssignStatement(
                    new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), WrappedObjectName),
                    new CodeVariableReferenceExpression(parameterName)
                ));

            declaration.Members.Add(codeConstructor);
        }



        /// <summary>
        /// 
        /// </summary>
        /// <param name="declaration"></param>
        /// <param name="properties">Tuple(string propertyName, Type propertyType, bool readOnly)</param>
        private static void AddInterfaceProperties(CodeTypeDeclaration declaration, IEnumerable<Tuple<string, Type, bool>> properties)
        {
            foreach (var property in properties)
            {
                string propertyName = property.Item1;
                Type propertyType = property.Item2;
                bool readOnly = property.Item3;

                string fieldName = CreateFieldName(propertyName);

                CodeTypeReference nullableType = new CodeTypeReference(
                        typeof(ExtendedNullable<>).FullName,
                        new CodeTypeReference[] { new CodeTypeReference(propertyType) }
                    );

                CodeMemberField codeField = new CodeMemberField();
                codeField.Name = fieldName;
                codeField.Type = nullableType;
                codeField.InitExpression = new CodeObjectCreateExpression(nullableType, new CodeExpression[] { });

                declaration.Members.Add(codeField);

                CodeMemberProperty codeProperty = new CodeMemberProperty();
                codeProperty.Attributes = MemberAttributes.Public | MemberAttributes.Final;
                codeProperty.Name = propertyName;
                codeProperty.HasGet = true;
                codeProperty.HasSet = !readOnly;
                codeProperty.Type = new CodeTypeReference(propertyType);


                codeProperty.GetStatements.Add(
                    new CodeConditionStatement(
                        new CodeBinaryOperatorExpression(
                            new CodePropertyReferenceExpression(
                                new CodeFieldReferenceExpression(
                                    new CodeThisReferenceExpression(),
                                    fieldName
                                ),
                                "HasValue"
                            ),
                            CodeBinaryOperatorType.IdentityEquality,
                            new CodePrimitiveExpression(true)
                        ),
                        new CodeStatement[] {
                            new CodeMethodReturnStatement(
                                new CodePropertyReferenceExpression(
                                    new CodeFieldReferenceExpression(
                                        new CodeThisReferenceExpression(),
                                        fieldName
                                    ),
                                    "Value"
                                )
                            )
                        },
                        new CodeStatement[] {
                            new CodeMethodReturnStatement(
                                new CodePropertyReferenceExpression(
                                    new CodeFieldReferenceExpression(
                                        new CodeThisReferenceExpression(),
                                        WrappedObjectName
                                    ),
                                    propertyName
                                )
                            )
                        }
                    ));


                if (!readOnly)
                {
                    codeProperty.SetStatements.Add(
                    new CodeAssignStatement(
                        new CodePropertyReferenceExpression(
                            new CodeFieldReferenceExpression(
                                new CodeThisReferenceExpression(),
                                fieldName
                            ),
                            "Value"
                        ),
                        new CodePropertySetValueReferenceExpression()
                    ));
                }

                declaration.Members.Add(codeProperty);
            }
        }




        /// <summary>
        /// 
        /// </summary>
        /// <param name="declaration"></param>
        /// <param name="properties">Tuple(string propertyName, Type propertyType, bool readOnly)</param>
        private static void AddMethods(CodeTypeDeclaration declaration, IEnumerable<Tuple<string, Type, bool>> properties)
        {
            CodeMemberProperty codeMemberProperty = new CodeMemberProperty();

            codeMemberProperty.Name = "WrappedData";
            codeMemberProperty.HasGet = true;
            codeMemberProperty.HasSet = false;
            codeMemberProperty.Type = new CodeTypeReference(typeof(IData));
            codeMemberProperty.Attributes = MemberAttributes.Public | MemberAttributes.Final;


            codeMemberProperty.GetStatements.Add(
                new CodeMethodReturnStatement(
                    new CodeFieldReferenceExpression(
                        new CodeThisReferenceExpression(),
                        WrappedObjectName
                    )
                ));

            declaration.Members.Add(codeMemberProperty);


            CodeMemberMethod codeMemberMethod = new CodeMemberMethod();
            codeMemberMethod.Name = "CommitData";
            codeMemberMethod.ReturnType = new CodeTypeReference(typeof(void));
            codeMemberMethod.Attributes = MemberAttributes.Public | MemberAttributes.Final;


            foreach (var property in properties)
            {
                string propertyName = property.Item1;
                bool readOnly = property.Item3;

                if (readOnly) continue;

                string fieldName = CreateFieldName(propertyName);

                List<CodeStatement> statements = new List<CodeStatement>();

                statements.Add(new CodeAssignStatement(
                        new CodePropertyReferenceExpression(
                            new CodeFieldReferenceExpression(
                                new CodeThisReferenceExpression(),
                                WrappedObjectName
                            ),
                            propertyName
                        ),
                        new CodePropertyReferenceExpression(
                            new CodeFieldReferenceExpression(
                                new CodeThisReferenceExpression(),
                                fieldName
                            ),
                            "Value"
                        )
                    ));


                codeMemberMethod.Statements.Add(new CodeConditionStatement(
                        new CodeBinaryOperatorExpression(
                            new CodePropertyReferenceExpression(
                                new CodeFieldReferenceExpression(
                                    new CodeThisReferenceExpression(),
                                    fieldName
                                ),
                                "HasValue"
                            ),
                            CodeBinaryOperatorType.IdentityEquality,
                            new CodePrimitiveExpression(true)
                        ),
                        statements.ToArray()
                    ));
            }

            declaration.Members.Add(codeMemberMethod);
        }




        private static string CreateFieldName(string propertyName)
        {
            return string.Format("_{0}Nullable", propertyName.ToLowerInvariant());
        }
    }
}
