/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.CodeDom;
using System.Collections.Generic;
using System.ComponentModel;
using System.Reflection;
using Composite.Core.Types;
using Composite.Data.DynamicTypes;


namespace Composite.Data.Foundation.CodeGeneration
{
    internal static class EmptyDataClassCodeGenerator
    {
        public static readonly string NamespaceName = "Composite.Data.GeneratedTypes";



        internal static void AddEmptyDataClassTypeCode(CodeGenerationBuilder codeGenerationBuilder, DataTypeDescriptor dataTypeDescriptor, Type baseClassType = null, CodeAttributeDeclaration codeAttributeDeclaration = null)
        {
            Type interfaceType = DataTypeTypesManager.GetDataType(dataTypeDescriptor);
            if (interfaceType == null) return;

            if (baseClassType == null) baseClassType = typeof(EmptyDataClassBase);

            CodeTypeDeclaration codeTypeDeclaration = CreateCodeTypeDeclaration(dataTypeDescriptor, baseClassType, codeAttributeDeclaration);

            codeGenerationBuilder.AddType(NamespaceName, codeTypeDeclaration);
        }



        internal static void AddAssemblyReferences(CodeGenerationBuilder codeGenerationBuilder, DataTypeDescriptor dataTypeDescriptor)
        {
            Type interfaceType = DataTypeTypesManager.GetDataType(dataTypeDescriptor);
            if (interfaceType == null) return;

            codeGenerationBuilder.AddReference(typeof(EmptyDataClassBase).Assembly);
            codeGenerationBuilder.AddReference(typeof(EditorBrowsableAttribute).Assembly);
            codeGenerationBuilder.AddReference(interfaceType.Assembly);

            if (!string.IsNullOrEmpty(dataTypeDescriptor.BuildNewHandlerTypeName))
            {
                Type buildeNewHandlerType = TypeManager.GetType(dataTypeDescriptor.BuildNewHandlerTypeName);
                codeGenerationBuilder.AddReference(buildeNewHandlerType.Assembly);
            }
        }



        internal static CodeTypeDeclaration CreateCodeTypeDeclaration(DataTypeDescriptor dataTypeDescriptor, Type baseClass, CodeAttributeDeclaration codeAttributeDeclaration)
        {
            Type interfaceType = DataTypeTypesManager.GetDataType(dataTypeDescriptor);

            string interfaceTypeFullName = interfaceType.FullName;
            CodeTypeDeclaration declaration = new CodeTypeDeclaration();

            declaration.Name = CreateClassName(interfaceTypeFullName);
            declaration.IsClass = true;
            declaration.TypeAttributes = TypeAttributes.Public | TypeAttributes.Sealed;
            declaration.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(SerializableAttribute))));
            declaration.CustomAttributes.Add(
                new CodeAttributeDeclaration(
                    new CodeTypeReference(typeof(EditorBrowsableAttribute)),
                    new CodeAttributeArgument(
                        new CodeFieldReferenceExpression(
                            new CodeTypeReferenceExpression(typeof(EditorBrowsableState)),
                            EditorBrowsableState.Never.ToString()
                        )
                    )
                )
            );

            if (baseClass != null)
            {
                declaration.BaseTypes.Add(baseClass);
            }
            declaration.BaseTypes.Add(new CodeTypeReference(interfaceType, CodeTypeReferenceOptions.GlobalReference));


            if (codeAttributeDeclaration != null)
            {
                declaration.CustomAttributes.Add(codeAttributeDeclaration);
            }


            AddConstructor(declaration);
            AddInterfaceProperties(declaration, dataTypeDescriptor.Fields);

            AddInterfaceTypeProperty(declaration, interfaceTypeFullName);

            return declaration;
        }



        private static void AddConstructor(CodeTypeDeclaration declaration)
        {
            CodeConstructor codeConstructor = new CodeConstructor();

            codeConstructor.Attributes = MemberAttributes.Public;

            declaration.Members.Add(codeConstructor);
        }



        private static void AddInterfaceProperties(CodeTypeDeclaration declaration, IEnumerable<DataFieldDescriptor> dataFieldDescriptors)
        {
            foreach (DataFieldDescriptor dataFieldDescriptor in dataFieldDescriptors)
            {
                string fieldName = CreateFieldName(dataFieldDescriptor);

                CodeMemberField codeField = new CodeMemberField(new CodeTypeReference(dataFieldDescriptor.InstanceType), fieldName);

                declaration.Members.Add(codeField);

                CodeMemberProperty property = new CodeMemberProperty();
                property.Attributes = MemberAttributes.Public | MemberAttributes.Final;
                property.Name = dataFieldDescriptor.Name;
                property.HasGet = true;
                property.HasSet = !dataFieldDescriptor.IsReadOnly;
                property.Type = new CodeTypeReference(dataFieldDescriptor.InstanceType);

                property.GetStatements.Add(
                        new CodeMethodReturnStatement(
                            new CodeFieldReferenceExpression(
                                new CodeThisReferenceExpression(),
                                fieldName
                            )
                        )
                    );


                if (!dataFieldDescriptor.IsReadOnly)
                {
                    property.SetStatements.Add(
                    new CodeAssignStatement(
                        new CodeFieldReferenceExpression(
                            new CodeThisReferenceExpression(),
                            fieldName
                        ),
                        new CodePropertySetValueReferenceExpression()
                    ));
                }

                declaration.Members.Add(property);
            }
        }


        private static void AddInterfaceTypeProperty(CodeTypeDeclaration declaration, string interfaceTypeFullName)
        {
            CodeMemberProperty codeMemberProperty = new CodeMemberProperty();
            codeMemberProperty.Name = "_InterfaceType";

            codeMemberProperty.Type = new CodeTypeReference(typeof(Type));
            codeMemberProperty.Attributes = MemberAttributes.Family | MemberAttributes.Override;
            codeMemberProperty.HasSet = false;
            codeMemberProperty.HasGet = true;
            codeMemberProperty.GetStatements.Add(
                    new CodeMethodReturnStatement(
                        new CodeTypeOfExpression(new CodeTypeReference(interfaceTypeFullName, CodeTypeReferenceOptions.GlobalReference))
                    )
                );

            declaration.Members.Add(codeMemberProperty);
        }



        internal static string GetEmptyClassTypeFullName(DataTypeDescriptor dataTypeDescriptor)
        {
            return NamespaceName + "." + CreateClassName(dataTypeDescriptor.GetFullInterfaceName());
        }



        private static string CreateClassName(string fullname)
        {
            return string.Format("{0}EmptyClass", fullname.Replace('.', '_').Replace('+', '_'));
        }



        private static string CreateFieldName(DataFieldDescriptor dataFieldDescriptor)
        {
            return string.Format("_{0}", dataFieldDescriptor.Name.ToLowerInvariant());
        }
    }    
}
