/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using Composite.Core.Collections.Generic;
using Composite.Data.Caching;
using Composite.Data.Foundation;
using Composite.Core.Logging;
using Composite.Core.Extensions;
using Composite.Core.Types;
using NullableGuid = Composite.Core.Types.ExtendedNullable<System.Guid>;
using Composite.Data.Types;

namespace Composite.Data
{
    /// <summary>
    /// Provides basic data access to IPage and IPageStructure data
    /// </summary>
    /// <exclude />
    // Made public for Base site in App_Code/Composite/BasicSearch.cs
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)]
    public static class PageManager
    {
        private static readonly string LogTitle = "PageManager";

        private class PageStructureRecord
        {
            public Guid ParentId { get; set; }
            public int LocalOrdering { get; set; }
        }

        private static readonly int PageCacheSize = 5000;
        private static readonly int PageStructureCacheSize = 3000;
        private static readonly int ChildrenCacheSize = 2000;
        private static readonly int PagePlaceholderCacheSize = 500;

        private static readonly Cache<string, ExtendedNullable<IPage>> _pageCache = new Cache<string, ExtendedNullable<IPage>>("Pages", PageCacheSize);
        private static readonly Cache<string, ExtendedNullable<PageStructureRecord>> _pageStructureCache = new Cache<string, ExtendedNullable<PageStructureRecord>>("Page structure", PageStructureCacheSize);
        private static readonly Cache<string, ReadOnlyCollection<Guid>> _childrenCache = new Cache<string, ReadOnlyCollection<Guid>>("Child pages", ChildrenCacheSize);
        private static readonly Cache<string, ReadOnlyCollection<IPagePlaceholderContent>> _placeholderCache = new Cache<string, ReadOnlyCollection<IPagePlaceholderContent>>("Page placeholders", PagePlaceholderCacheSize);

        static PageManager()
        {
            SubscribeToEvents();
        }

        #region Public methods


        /// <exclude />
        public static IPage GetPageById(Guid id)
        {
            return GetPageById(id, false);
        }


        /// <exclude />
        public static IPage GetPageById(Guid id, bool readonlyValue)
        {
            IPage result;

            string cacheKey = GetCacheKey<IPage>(id);
            var cachedValue = _pageCache.Get(cacheKey);

            if (cachedValue != null)
            {
                result = cachedValue.Value;
            }
            else
            {
                result = (from page in DataFacade.GetData<IPage>(false)
                          where page.Id == id
                          select page).FirstOrDefault();

                _pageCache.Add(cacheKey, new ExtendedNullable<IPage> { Value = result });
            }

            if (result == null)
            {
                return null;
            }

            return readonlyValue ? result : CreateWrapper(result);
        }


        /// <exclude />
        [Obsolete("Use GetParentId(..)", true)]
        public static Guid GetParentID(Guid pageId)
        {
            return GetParentId(pageId);
        }


        /// <exclude />
        public static Guid GetParentId(Guid pageId)
        {
            PageStructureRecord pageStructure = GetPageStructureRecord(pageId);
            return pageStructure != null ? pageStructure.ParentId : Guid.Empty;
        }


        /// <exclude />
        public static int GetLocalOrdering(Guid pageId)
        {
            PageStructureRecord pageStructure = GetPageStructureRecord(pageId);
            return pageStructure != null ? pageStructure.LocalOrdering : 0;
        }



        /// <summary>
        /// 
        /// </summary>
        /// <param name="parentPageId">Empty guild will yield all root pages</param>
        /// <returns></returns>
        public static ReadOnlyCollection<Guid> GetChildrenIDs(Guid parentPageId)
        {
            var cacheKey = GetCacheKey<IPageStructure>(parentPageId);
            var cachedValue = _childrenCache.Get(cacheKey);

            if (cachedValue != null)
            {
                return cachedValue;
            }

            List<Guid> children = (from ps in DataFacade.GetData<IPageStructure>()
                                   where ps.ParentId == parentPageId
                                   orderby ps.LocalOrdering
                                   select ps.Id).ToList();

            var readonlyList = new ReadOnlyCollection<Guid>(children);

            _childrenCache.Add(cacheKey, readonlyList);

            return readonlyList;
        }



        /// <exclude />
        public static ReadOnlyCollection<IPagePlaceholderContent> GetPlaceholderContent(Guid pageId)
        {
            string cacheKey = GetCacheKey<IPagePlaceholderContent>(pageId);
            var cachedValue = _placeholderCache.Get(cacheKey);
            if (cachedValue != null)
            {
                return cachedValue;
            }

            var list = DataFacade.GetData<IPagePlaceholderContent>(f => f.PageId == pageId).ToList();

            var readonlyList = new ReadOnlyCollection<IPagePlaceholderContent>(list);

            _placeholderCache.Add(cacheKey, readonlyList);

            return readonlyList;
        }

        #endregion Public

        #region Private

        private static PageStructureRecord GetPageStructureRecord(Guid pageId)
        {
            string cacheKey = GetCacheKey<IPageStructure>(pageId);
            ExtendedNullable<PageStructureRecord> cachedValue = _pageStructureCache.Get(cacheKey);

            if (cachedValue != null)
            {
                Verify.That(cachedValue.HasValue, "Incorrect usage of cache.");
                return cachedValue.Value;
            }

            PageStructureRecord result =
                         (from ps in DataFacade.GetData<IPageStructure>(false)
                          where ps.Id == pageId
                          select new PageStructureRecord
                          {
                              ParentId = ps.ParentId,
                              LocalOrdering = ps.LocalOrdering
                          }).FirstOrDefault();

            if (result == null)
            {
                LoggingService.LogWarning(LogTitle,
                    "No IPageStructure entries found for Page with Id '{0}'".FormatWith(pageId));
            }

            _pageStructureCache.Add(cacheKey, result);

            return result;
        }

        private static string GetCacheKey<T>(Guid id)
        {
            string localizationInfo = LocalizationScopeManager.MapByType(typeof(T)).ToString();
            string dataScope = DataScopeManager.MapByType(typeof(T)).Name;
            return id + dataScope + localizationInfo;
        }

        private static string GetCacheKey(Guid id, DataSourceId dataSourceId)
        {
            string localizationInfo = dataSourceId.LocaleScope.ToString();
            string dataScope = dataSourceId.DataScopeIdentifier.Name;
            return id + dataScope + localizationInfo;
        }

        private static void OnPageStoreChanged(object sender, StoreEventArgs storeEventArgs)
        {
            if (!storeEventArgs.DataEventsFired)
            {
                _pageCache.Clear();
            }
        }

        private static void OnPageChanged(object sender, DataEventArgs args)
        {
            IPage page = args.Data as IPage;
            if (page == null)
            {
                return;
            }

            _pageCache.Remove(GetCacheKey(page.Id, page.DataSourceId));
        }


        private static void OnPagePlaceholderStoreChanged(object sender, StoreEventArgs storeEventArgs)
        {
            if (!storeEventArgs.DataEventsFired)
            {
                _placeholderCache.Clear();
            }
        }

        private static void OnPagePlaceholderChanged(object sender, DataEventArgs args)
        {
            var placeHolder = args.Data as IPagePlaceholderContent;
            if (placeHolder == null)
            {
                return;
            }

            _placeholderCache.Remove(GetCacheKey(placeHolder.PageId, placeHolder.DataSourceId));
        }


        private static void OnPageStructureStoreChanged(object sender, StoreEventArgs storeEventArgs)
        {
            if (!storeEventArgs.DataEventsFired)
            {
                _pageStructureCache.Clear();
                _childrenCache.Clear();
            }
        }

        private static void OnPageStructureChanged(object sender, DataEventArgs args)
        {
            var ps = args.Data as IPageStructure;
            if (ps == null)
            {
                return;
            }

            _pageStructureCache.Remove(GetCacheKey(ps.Id, ps.DataSourceId));
            _childrenCache.Remove(GetCacheKey(ps.ParentId, ps.DataSourceId));
        }

        private static IPage CreateWrapper(IPage page)
        {
            return DataWrappingFacade.Wrap(page);
        }

        private static void SubscribeToEvents()
        {
            DataEventSystemFacade.SubscribeToDataDeleted<IPagePlaceholderContent>(OnPagePlaceholderChanged, true);
            DataEventSystemFacade.SubscribeToDataAfterUpdate<IPagePlaceholderContent>(OnPagePlaceholderChanged, true);
            DataEventSystemFacade.SubscribeToDataAfterAdd<IPagePlaceholderContent>(OnPagePlaceholderChanged, true);
            DataEventSystemFacade.SubscribeToStoreChanged<IPagePlaceholderContent>(OnPagePlaceholderStoreChanged, true);

            DataEventSystemFacade.SubscribeToDataAfterUpdate<IPage>(OnPageChanged, true);
            DataEventSystemFacade.SubscribeToDataAfterAdd<IPage>(OnPageChanged, true);
            DataEventSystemFacade.SubscribeToDataDeleted<IPage>(OnPageChanged, true);
            DataEventSystemFacade.SubscribeToStoreChanged<IPage>(OnPageStoreChanged, true);

            DataEventSystemFacade.SubscribeToDataAfterAdd<IPageStructure>(OnPageStructureChanged, true);
            DataEventSystemFacade.SubscribeToDataAfterUpdate<IPageStructure>(OnPageStructureChanged, true);
            DataEventSystemFacade.SubscribeToDataDeleted<IPageStructure>(OnPageStructureChanged, true);
            DataEventSystemFacade.SubscribeToStoreChanged<IPageStructure>(OnPageStructureStoreChanged, true);
            
        }

        #endregion Private
    }
}
