/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Reflection;
using System.Threading;
using Composite.Core.Collections.Generic;
using Composite.Core.IO;
using Composite.Core.Types;
using Composite.Data.Streams;
using System.IO;

namespace Composite.Data.Plugins.DataProvider.Streams
{
	internal class FileChangeNotificator
	{
        private static readonly object _syncRoot = new object();
        private static C1FileSystemWatcher _fileWatcher;

        private static int _counter;

        // We're holding only weak reference to subscriber objects, in order to avoid memory leaks
        private static readonly Hashtable<string, ReadOnlyCollection<Pair<MethodInfo, WeakReference>>> _subscribers = new Hashtable<string, ReadOnlyCollection<Pair<MethodInfo, WeakReference>>>();


        private static void EnsureInitialization()
        {
            if (_fileWatcher != null)
            {
                return;
            }

            lock (_syncRoot)
            {
                if (_fileWatcher != null)
                {
                    return;
                }

                _fileWatcher = new C1FileSystemWatcher(AppDomain.CurrentDomain.BaseDirectory);
                _fileWatcher.Created += FileWatcher_Created;
                _fileWatcher.Changed += FileWatcher_Changed;
                _fileWatcher.Deleted += FileWatcher_Deleted;

                _fileWatcher.IncludeSubdirectories = true;
                _fileWatcher.NotifyFilter = NotifyFilters.FileName | NotifyFilters.LastWrite;

                _fileWatcher.EnableRaisingEvents = true;
            }
        }


        static void FileWatcher_Deleted(object sender, FileSystemEventArgs e)
        {
            FireFileChangedEvent(e.FullPath, FileChangeType.Deleted);
        }

        static void FileWatcher_Changed(object sender, FileSystemEventArgs e)
        {
            FileChangeType changeType = e.ChangeType == WatcherChangeTypes.Renamed
                                            ? FileChangeType.Renamed
                                            : FileChangeType.Modified;
            FireFileChangedEvent(e.FullPath, changeType);
        }

        static void FileWatcher_Created(object sender, FileSystemEventArgs e)
        {
            // Do nothing...
        }

        private static void FireFileChangedEvent(string filePath, FileChangeType changeType)
        {
            filePath = filePath.ToLowerInvariant();

            ReadOnlyCollection<Pair<MethodInfo, WeakReference>> weakInvocationList;

            if (!_subscribers.TryGetValue(filePath.ToLowerInvariant(), out weakInvocationList))
            {
                return;
            }

            var parameters = new object[] { filePath, changeType };

            foreach (var callInfo in weakInvocationList)
            {
                if (callInfo.Second == null) // Call to a static method
                {
                    callInfo.First.Invoke(null, parameters);
                }
                else
                {
                    object target = callInfo.Second.Target;
                    if (target != null) // Checking if object is alive
                    {
                        callInfo.First.Invoke(target, parameters);
                    }
                }
            }
        }


        public static void Subscribe(FileSystemFileBase file, OnFileChangedDelegate handler)
        {
            Verify.ArgumentNotNull(file, "file");

            Subscribe(file.SystemPath, handler);
        }


	    public static void Subscribe(string filePath, OnFileChangedDelegate handler)
        {
            Verify.ArgumentNotNullOrEmpty(filePath, "filePath");
            Verify.ArgumentNotNull(handler, "handler");

            EnsureInitialization();

            int counterValue = Interlocked.Increment(ref _counter);
            if (counterValue % 100 == 0)
            {
                ClearDeadReferences();
            }

            var weakInvocationList = new List<Pair<MethodInfo, WeakReference>>();
            foreach (Delegate func in handler.GetInvocationList())
            {
                var targetObject = func.Target;
                if (targetObject == null)
                {
                    weakInvocationList.Add(new Pair<MethodInfo, WeakReference>(func.Method, null));
                }
                else
                {
                    weakInvocationList.Add(new Pair<MethodInfo, WeakReference>(func.Method, new WeakReference(handler.Target)));
                }
            }

            string key = filePath.ToLowerInvariant();
            lock (_syncRoot)
            {
                if (_subscribers.ContainsKey(key))
                {
                    ReadOnlyCollection<Pair<MethodInfo, WeakReference>> oldList = _subscribers[key];

                    var newList = new ReadOnlyCollection<Pair<MethodInfo, WeakReference>>(
                        new List<Pair<MethodInfo, WeakReference>>(oldList.Concat(weakInvocationList)));

                    _subscribers[key] = newList;
                }
                else
                {
                    _subscribers.Add(key, new ReadOnlyCollection<Pair<MethodInfo, WeakReference>>(weakInvocationList));
                }
            }
        }

        private static void ClearDeadReferences()
        {
            lock(_syncRoot)
            {
                ICollection<string> keys = _subscribers.GetKeys();

                foreach(string key in keys)
                {
                    ReadOnlyCollection<Pair<MethodInfo, WeakReference>> currentList = _subscribers[key];

                    int countOfAlive = currentList.Count(pair => pair.Second == null || pair.Second.IsAlive);
                    if(countOfAlive == 0)
                    {
                        _subscribers.Remove(key);
                        continue;
                    }

                    if (countOfAlive == currentList.Count)
                    {
                        continue;
                    }

                    var newList = new List<Pair<MethodInfo, WeakReference>>(
                        currentList.Where(pair => pair.Second == null || pair.Second.IsAlive));

                    _subscribers[key] = new ReadOnlyCollection<Pair<MethodInfo, WeakReference>>(newList);
                }
            }
        }
	}
}
