/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Transactions;
using Composite.Core.Caching;
using Composite.Data.Foundation;
using Composite.Data.ProcessControlled.ProcessControllers.DummyProcessControllers;
using Composite.C1Console.Elements;
using Composite.C1Console.Events;
using Composite.Core.Logging;
using Composite.Data.Transactions;
using Composite.Core.Types;
using Composite.Core.Linq;


namespace Composite.Data.ProcessControlled
{
    /// <summary>    
    /// </summary>
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)] 
    public static class ProcessControllerFacade
    {
        private static Dictionary<Type, Type> _controlledTypes = new Dictionary<Type, Type>();

        private static Dictionary<Type, IProcessController> _processControllers;
        private static Dictionary<Type, List<IPublishControlledAuxiliary>> _publishControlledAuxiliaries = new Dictionary<Type, List<IPublishControlledAuxiliary>>();

        private static object _lock = new object();


        static ProcessControllerFacade()
        {
            GlobalEventSystemFacade.SubscribeToFlushEvent(OnFlushEvent);

            _controlledTypes.Add(typeof(IPublishControlled), typeof(IPublishProcessController));
            _controlledTypes.Add(typeof(ILocalizedControlled), typeof(ILocalizeProcessController));
        }




        /// <summary>
        /// This method will delete the data, and its if the exists its versions and publications
        /// </summary>
        /// <param name="data"></param>
        public static void FullDelete(IData data)
        {
            using (TransactionScope transactionScope = TransactionsFacade.CreateNewScope())
            {
                using (DataScope adminDataScope = new DataScope(DataScopeIdentifier.Administrated))
                {
                    if (data is IPublishControlled)
                    {
                        using (DataScope dataScope = new DataScope(DataScopeIdentifier.Public))
                        {                        
                            IEnumerable<IData> datasDelete = DataFacade.GetDataFromOtherScope(data, DataScopeIdentifier.Public).Evaluate();
                        
                            DataFacade.Delete(datasDelete, CascadeDeleteType.Disable);
                        }
                    }

                    DataFacade.Delete(data);
                }

                transactionScope.Complete();
            }
        }



        /// <summary>
        /// This method return the types that implements the same IProcessControlled interfaces as
        /// the targetType.
        /// </summary>
        /// <param name="targetType"></param>
        /// <param name="types"></param>
        /// <returns></returns>
        public static IEnumerable<Type> GetProcessControlledTypes(Type targetType, IEnumerable<Type> types)
        {
            if (targetType == null) throw new ArgumentNullException("targetType");
            if (types == null) throw new ArgumentNullException("types");

            List<Type> compatibleTypes = new List<Type>();

            foreach (Type type in types)
            {
                foreach (Type processControledType in _controlledTypes.Keys)
                {
                    if ((compatibleTypes.Contains(type) == false) &&
                        (processControledType.IsAssignableFrom(targetType)) &&
                        (processControledType.IsAssignableFrom(type)))
                    {
                        compatibleTypes.Add(type);
                    }
                }
            }

            return compatibleTypes;
        }




        /// <exclude />
        public static List<ElementAction> GetActions(IData data, Type elementProviderType)
        {
            if (data == null) throw new ArgumentNullException("data");
            if (elementProviderType == null) throw new ArgumentNullException("elementProviderType");

            List<ElementAction> actions = new List<ElementAction>();

            Dictionary<Type, Type> processControllerTypes = ProcessControllerRegistry.GetProcessControllerTypes(data.DataSourceId.InterfaceType);

            foreach (Type processControllerType in processControllerTypes.Values)
            {
                IProcessController processController = _processControllers[processControllerType];

                List<ElementAction> acts = processController.GetActions(data, elementProviderType);

                if (acts != null)
                {
                    actions.AddRange(acts);
                }
            }

            return actions;
        }


        #region IPublishControlled methods
        /// <exclude />
        public static void SetStartStatus(IData data)
        {
            if (data == null) throw new ArgumentNullException("data");

            IPublishProcessController controller = GetProcessController<IPublishProcessController>(data.DataSourceId.InterfaceType);

            if (controller == null) throw new ArgumentException(string.Format("The type '{0}' is not registred process controller", data.DataSourceId.InterfaceType));

            controller.SetStartStatus(data);
        }



        /// <exclude />
        public static IDictionary<string, string> GetValidTransitions(IData data)
        {
            if (data == null) throw new ArgumentNullException("data");

            IPublishProcessController controller = GetProcessController<IPublishProcessController>(data.DataSourceId.InterfaceType);

            if (controller == null) throw new ArgumentException(string.Format("The type '{0}' is not registred process controller", data.DataSourceId.InterfaceType));

            return controller.GetValidTransitions(data);
        }



        /// <exclude />
        public static void ValidateTransition(IData data, string status)
        {
            if (data == null) throw new ArgumentNullException("data");

            IPublishProcessController controller = GetProcessController<IPublishProcessController>(data.DataSourceId.InterfaceType);

            if (controller == null) throw new ArgumentException(string.Format("The type '{0}' is not registred process controller", data.DataSourceId.InterfaceType));

            controller.ValidateTransition(data, status);
        }
        #endregion



        /// <exclude />
        public static Type GetProcessControllerType<T>(IData data)
            where T : class, IProcessController
        {
            if (data == null) throw new ArgumentNullException("data");

            T controller = GetProcessController<T>(data.DataSourceId.InterfaceType);

            if (controller == null) throw new ArgumentException(string.Format("The type '{0}' is not registred process controller", data.DataSourceId.InterfaceType));

            return controller.GetType();
        }



        /// <exclude />
        public static IDisposable NoProcessControllers
        {
            get
            {
                return new NoProcessControllersDisposable();
            }
        }


        private static void OnAfterDataUpdated(object sender, DataEventArgs dataEventArgs)
        {
            if (GetProcessControllersCounter() > 0) return;

            if (dataEventArgs.Data is IPublishControlled)
            {
                IPublishProcessController controller = GetProcessController<IPublishProcessController>(dataEventArgs.Data.DataSourceId.InterfaceType);

                if (controller != null)
                {
                    bool callAuxilaries = controller.OnAfterDataUpdated(dataEventArgs.Data);

                    if (callAuxilaries)
                    {
                        IEnumerable<IPublishControlledAuxiliary> publishControlledAuxiliaries = GetPublishControlledAuxiliaries(dataEventArgs.DataType);

                        foreach (IPublishControlledAuxiliary publishControlledAuxiliary in publishControlledAuxiliaries)
                        {
                            publishControlledAuxiliary.OnAfterDataUpdated(dataEventArgs.Data);
                        }
                    }
                }
            }
        }



        private static void OnDataBuildNew(object sender, DataEventArgs dataEventArgs)
        {
            if (GetProcessControllersCounter() > 0) return;

            if (dataEventArgs.Data is IPublishControlled)
            {
                IPublishProcessController controller = GetProcessController<IPublishProcessController>(dataEventArgs.Data.DataSourceId.InterfaceType);

                if (controller != null)
                {
                    bool callAuxilaries = controller.OnAfterBuildNew(dataEventArgs.Data);

                    if (callAuxilaries)
                    {
                        IEnumerable<IPublishControlledAuxiliary> publishControlledAuxiliaries = GetPublishControlledAuxiliaries(dataEventArgs.DataType);

                        foreach (IPublishControlledAuxiliary publishControlledAuxiliary in publishControlledAuxiliaries)
                        {
                            publishControlledAuxiliary.OnAfterDataUpdated(dataEventArgs.Data);
                        }
                    }
                }
            }

            if (dataEventArgs.Data is ILocalizedControlled)
            {
                ILocalizeProcessController controller = GetProcessController<ILocalizeProcessController>(dataEventArgs.Data.DataSourceId.InterfaceType);

                if (controller != null)
                {
                    controller.OnAfterBuildNew(dataEventArgs.Data);
                }
            }
        }



        private static T GetProcessController<T>(Type dataType)
            where T : class, IProcessController
        {
            Dictionary<Type, Type> processControllerTypes = ProcessControllerRegistry.GetProcessControllerTypes(dataType);

            Type type;
            if (processControllerTypes.TryGetValue(typeof(T), out type) == false)
            {
                return null;
            }

            return (T)_processControllers[type];
        }



        private static IEnumerable<IPublishControlledAuxiliary> GetPublishControlledAuxiliaries(Type dataType)
        {
            List<IPublishControlledAuxiliary> publishControlledAuxiliaries;

            if (_publishControlledAuxiliaries.TryGetValue(dataType, out publishControlledAuxiliaries) == false)
            {
                List<PublishControlledAuxiliaryAttribute> attributes = dataType.GetCustomAttributesRecursively<PublishControlledAuxiliaryAttribute>().ToList();

                publishControlledAuxiliaries = new List<IPublishControlledAuxiliary>();

                foreach (PublishControlledAuxiliaryAttribute attribute in attributes)
                {
                    if (attribute.PublishControlledAuxiliaryType == null) throw new InvalidOperationException(string.Format("The PublishControlledAuxiliaryType may not be null on the {0}", typeof(PublishControlledAuxiliaryAttribute)));
                    if (typeof(IPublishControlledAuxiliary).IsAssignableFrom(attribute.PublishControlledAuxiliaryType) == false) throw new InvalidOperationException(string.Format("The {0} does not inheret the interface {1} the {2}", attribute.PublishControlledAuxiliaryType, typeof(IPublishControlledAuxiliary), typeof(PublishControlledAuxiliaryAttribute)));

                    ConstructorInfo constructorInfo = attribute.PublishControlledAuxiliaryType.GetConstructor(new Type[] { });
                    if (constructorInfo == null) throw new InvalidOperationException(string.Format("The type {0} used by the {1} does not have a default contructor", attribute.PublishControlledAuxiliaryType, typeof(PublishControlledAuxiliaryAttribute)));

                    IPublishControlledAuxiliary publishControlledAuxiliary = (IPublishControlledAuxiliary)Activator.CreateInstance(attribute.PublishControlledAuxiliaryType, new object[] { });

                    publishControlledAuxiliaries.Add(publishControlledAuxiliary);
                }

                _publishControlledAuxiliaries.Add(dataType, publishControlledAuxiliaries);
            }

            foreach (IPublishControlledAuxiliary publishControlledAuxiliary in publishControlledAuxiliaries)
            {
                yield return publishControlledAuxiliary;
            }
        }



        private static Type MapProcessControlledTypeToProcessControllerType(Type processControlledType)
        {
            return _controlledTypes[processControlledType];
        }



        internal static void Initialize_PostDataTypes()
        {
            if (RuntimeInformation.IsDebugBuild)
            {
                GlobalInitializerFacade.ValidateIsOnlyCalledFromGlobalInitializerFacade(new StackTrace());
            }

            DoInitialize();
        }



        private static void DoInitialize()
        {
            _processControllers = new Dictionary<Type, IProcessController>();

            foreach (Type processControllerType in ProcessControllerRegistry.ProcessControllerTypes)
            {
                IProcessController processController = (IProcessController)Activator.CreateInstance(processControllerType);

                _processControllers.Add(processControllerType, processController);
            }


            foreach (Type interfaceType in ProcessControllerRegistry.DataTypesWithProcessControllers)
            {
                Dictionary<Type, Type> processControllerTypes = ProcessControllerRegistry.GetProcessControllerTypes(interfaceType);
            }

            if (RuntimeInformation.IsUnittest == false)
            {
                foreach (Type dataType in DataFacade.GetAllInterfaces())
                {
                    Dictionary<Type, Type> processControllerTypes = ProcessControllerRegistry.GetProcessControllerTypes(dataType);

                    foreach (var kvp in _controlledTypes)
                    {
                        if (kvp.Key.IsAssignableFrom(dataType))
                        {
                            if (processControllerTypes.ContainsKey(kvp.Value) == false)
                            {
                                throw new InvalidOperationException(string.Format("The data type {0} is inheriting the interface {1} but has not been assigned a {2} process controller", dataType, kvp.Key, kvp.Value));
                            }
                        }
                    }
                }
            }

            DataEventSystemFacade.SubscribeToDataAfterUpdate<IProcessControlled>(OnAfterDataUpdated, false);
            DataEventSystemFacade.SubscribeToDataAfterBuildNew<IProcessControlled>(OnDataBuildNew, false);
        }



        private static void Flush()
        {
            _processControllers = null;
            _publishControlledAuxiliaries = new Dictionary<Type, List<IPublishControlledAuxiliary>>();
        }



        private static void OnFlushEvent(FlushEventArgs args)
        {
            Flush();
        }



        private static void PushProcessControllers()
        {
            ProcessControllersCounter.Counter++;
        }



        private static void PopProcessControllers()
        {
            ProcessControllersCounter.Counter--;
        }



        private static int GetProcessControllersCounter()
        {
            return ProcessControllersCounter.Counter;
        }



        private sealed class NoProcessControllersDisposable : IDisposable
        {
            public NoProcessControllersDisposable()
            {
                ProcessControllerFacade.PushProcessControllers();
            }

            public void Dispose()
            {
                ProcessControllerFacade.PopProcessControllers();
            }
        }



        private sealed class CounterContainer
        {
            public CounterContainer()
            {
                this.Counter = 0;
            }

            public int Counter { get; set; }
        }


        private static CounterContainer ProcessControllersCounter
        {
            get
            {
                return RequestLifetimeCache.GetCachedOrNew<CounterContainer>("ProcessControllerFacade:ProcessControllersCounter");
            }
        }
    }
}
