/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.CodeDom;
using System.ComponentModel;
using System.Reflection;
using System.Xml.Linq;
using Composite.Core.Types;
using Composite.Data;
using Composite.Data.DynamicTypes;


namespace Composite.Plugins.Data.DataProviders.XmlDataProvider.CodeGeneration
{
    /// <summary>
    /// Creates a implementation of DataProviderHelperBase
    /// It does NOT depend on the data type to exist
    /// </summary>
    internal sealed class DataProviderHelperClassGenerator
    {
        private const string WrapperClassConstructorFieldName = "_wrapperClassConstructor";
        private const string DataIdClassConstructorFieldName = "_idClassConstructor";

        private readonly string _helperClassName;
        private readonly string _wrapperClassName;
        private readonly string _dataIdClassName;
        private string _interfaceName;

        private readonly DataTypeDescriptor _dataTypeDescriptor;


        public DataProviderHelperClassGenerator(
                string helperClassName,
                string wrapperClassName,
                string dataIdClassName,
                DataTypeDescriptor dataTypeDescriptor)
        {
            _helperClassName = helperClassName;
            _wrapperClassName = wrapperClassName;
            _dataIdClassName = dataIdClassName;
            _dataTypeDescriptor = dataTypeDescriptor;
        }



        public CodeTypeDeclaration CreateClass()
        {
            CodeTypeDeclaration declaration = new CodeTypeDeclaration(_helperClassName);

            declaration.IsClass = true;
            declaration.TypeAttributes = TypeAttributes.Public | TypeAttributes.Sealed;
            declaration.BaseTypes.Add(typeof(DataProviderHelperBase));
            declaration.CustomAttributes.Add(
                new CodeAttributeDeclaration(
                    new CodeTypeReference(typeof(EditorBrowsableAttribute)),
                    new CodeAttributeArgument(
                        new CodeFieldReferenceExpression(
                            new CodeTypeReferenceExpression(typeof(EditorBrowsableState)),
                            EditorBrowsableState.Never.ToString()
                        )
                    )
                )
            );

            AddConstructor(declaration);
            AddInterfaceTypeProperty(declaration);
            AddDataIdTypeProperty(declaration);
            AddCreateDataIdFunctionMethod(declaration);
            AddValidateDataTypeMethod(declaration);
            AddCreateNewElementMethod(declaration);

            return declaration;
        }



        private void AddConstructor(CodeTypeDeclaration declaration)
        {
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public | MemberAttributes.Final;

            constructor.Statements.Add(new CodeAssignStatement(
                    new CodeFieldReferenceExpression(
                        new CodeThisReferenceExpression(),
                        WrapperClassConstructorFieldName
                    ),
                    new CodeArrayIndexerExpression(
                        new CodeMethodInvokeExpression(
                            new CodeTypeOfExpression(
                                _wrapperClassName
                            ),
                            "GetConstructors",
                            new CodeExpression[] {
                                new CodeBinaryOperatorExpression(
                                    new CodeFieldReferenceExpression(
                                        new CodeTypeReferenceExpression(typeof(BindingFlags)),
                                        "Instance"
                                    ),
                                    CodeBinaryOperatorType.BitwiseOr,
                                    new CodeFieldReferenceExpression(
                                        new CodeTypeReferenceExpression(typeof(BindingFlags)),
                                        "Public"
                                    )
                                )        
                            }
                        ),
                        new CodeExpression[] { new CodePrimitiveExpression(0) }
                    )
                ));


            constructor.Statements.Add(new CodeAssignStatement(
                    new CodeFieldReferenceExpression(
                        new CodeThisReferenceExpression(),
                        DataIdClassConstructorFieldName
                    ),
                    new CodeArrayIndexerExpression(
                        new CodeMethodInvokeExpression(
                            new CodeTypeOfExpression(
                                _dataIdClassName
                            ),
                            "GetConstructors",
                            new CodeExpression[] {
                                new CodeBinaryOperatorExpression(
                                    new CodeFieldReferenceExpression(
                                        new CodeTypeReferenceExpression(typeof(BindingFlags)),
                                        "Instance"
                                    ),
                                    CodeBinaryOperatorType.BitwiseOr,
                                    new CodeFieldReferenceExpression(
                                        new CodeTypeReferenceExpression(typeof(BindingFlags)),
                                        "Public"
                                    )
                                )        
                            }
                        ),
                        new CodeExpression[] { new CodePrimitiveExpression(1) }
                    )
                ));


            declaration.Members.Add(constructor);
        }



        private void AddInterfaceTypeProperty(CodeTypeDeclaration declaration)
        {
            CodeMemberProperty property = new CodeMemberProperty();
            property.Name = "_InterfaceType";
            property.HasGet = true;
            property.HasSet = false;
            property.Type = new CodeTypeReference(typeof(Type));
            property.Attributes = MemberAttributes.Public | MemberAttributes.Override;

            property.GetStatements.Add(new CodeMethodReturnStatement(
                    new CodeTypeOfExpression(_dataTypeDescriptor.GetFullInterfaceName())
                ));


            declaration.Members.Add(property);
        }



        private void AddDataIdTypeProperty(CodeTypeDeclaration declaration)
        {
            CodeMemberProperty property = new CodeMemberProperty();
            property.Name = "_DataIdType";
            property.HasGet = true;
            property.HasSet = false;
            property.Type = new CodeTypeReference(typeof(Type));
            property.Attributes = MemberAttributes.Public | MemberAttributes.Override;

            property.GetStatements.Add(new CodeMethodReturnStatement(
                    new CodeTypeOfExpression(_dataIdClassName)
                ));


            declaration.Members.Add(property);
        }



        private void AddCreateDataIdFunctionMethod(CodeTypeDeclaration declaration)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Name = "CreateDataId";
            method.ReturnType = new CodeTypeReference(typeof(IDataId));
            method.Parameters.Add(new CodeParameterDeclarationExpression(typeof(XElement), "xElement"));
            method.Attributes = MemberAttributes.Public | MemberAttributes.Override;

            method.Statements.Add(new CodeMethodReturnStatement(
                new CodeObjectCreateExpression(this._dataIdClassName,
                new CodeArgumentReferenceExpression("xElement"))));

            declaration.Members.Add(method);
        }


        private static void AddValidateDataTypeMethod(CodeTypeDeclaration declaration)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Name = "ValidateDataType";
            method.ReturnType = new CodeTypeReference(typeof(void));
            method.Parameters.Add(new CodeParameterDeclarationExpression(typeof(IData), "data"));
            method.Attributes = MemberAttributes.Public | MemberAttributes.Override;



            declaration.Members.Add(method);
        }



        private void AddCreateNewElementMethod(CodeTypeDeclaration declaration)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Name = "CreateNewElement";

            CodeTypeParameter genericParameter = new CodeTypeParameter("T");
            genericParameter.HasConstructorConstraint = false;

            method.TypeParameters.Add(genericParameter);

            method.ReturnType = new CodeTypeReference(genericParameter);
            method.ImplementationTypes.Add(typeof(IXmlDataProviderHelper));

            method.Parameters.Add(new CodeParameterDeclarationExpression(typeof(IData), "data"));
            CodeParameterDeclarationExpression parameter = new CodeParameterDeclarationExpression(typeof(XElement), "newElement");
            parameter.Direction = FieldDirection.Out;
            method.Parameters.Add(parameter);
            method.Parameters.Add(new CodeParameterDeclarationExpression(typeof(string), "elementName"));
            method.Parameters.Add(new CodeParameterDeclarationExpression(typeof(string), "providerName"));

            method.Attributes = MemberAttributes.Public | MemberAttributes.Override;


            method.Statements.Add(new CodeVariableDeclarationStatement(
                InterfaceName,
                "xmlData"));


            method.Statements.Add(new CodeTryCatchFinallyStatement(
                    new CodeStatement[] {
                            new CodeAssignStatement(
                                new CodeVariableReferenceExpression("xmlData"),
                                new CodeCastExpression(
                                    InterfaceName,
                                    new CodeVariableReferenceExpression("data"))
                            )
                    },
                    new CodeCatchClause[] {
                         new CodeCatchClause("e", new CodeTypeReference(typeof(Exception)), new CodeStatement[] {
                            new CodeThrowExceptionStatement(
                                new CodeObjectCreateExpression(
                                    new CodeTypeReference(typeof(ArgumentException)),
                                    new CodeExpression[] {
                                        new CodeMethodInvokeExpression(
                                            new CodeTypeReferenceExpression(typeof(string)), 
                                            "Format",
                                            new CodeExpression[] {
                                                new CodePrimitiveExpression("The type ({0}) of the given data parameter does not match the type {1}"),
                                                new CodeMethodInvokeExpression(
                                                    new CodeVariableReferenceExpression("data"),
                                                    "GetType",
                                                    new CodeExpression[] {}
                                                ),
                                                new CodeTypeOfExpression(InterfaceName)
                                            }
                                        ),
                                        new CodeVariableReferenceExpression("e")
                                    }                                    
                                )
                            )
                        })
                    }
                ));


            const string newElementVariableName = "newElement";

            method.Statements.Add(new CodeAssignStatement(
                new CodeVariableReferenceExpression(newElementVariableName),
                new CodeObjectCreateExpression(
                        typeof(XElement),
                        new CodeVariableReferenceExpression("elementName")
                    )
            ));



            foreach (DataFieldDescriptor field in _dataTypeDescriptor.Fields)
            {
                method.Statements.Add(new CodeCommentStatement(string.Format("Interface Property {0}", field.Name)));

                CodeMethodInvokeExpression codeExpression = new CodeMethodInvokeExpression(
                            new CodeVariableReferenceExpression(newElementVariableName),
                            "Add",
                            new CodeExpression[] {
                                    new CodeObjectCreateExpression(
                                        typeof(XAttribute),
                                        new CodeExpression[] {
                                            new CodePrimitiveExpression(field.Name),
                                            new CodePropertyReferenceExpression(
                                                new CodeVariableReferenceExpression("xmlData"),
                                                field.Name
                                            )
                                        }
                                    )
                                }
                        );

                if ((field.InstanceType.IsGenericType) &&
                    (field.InstanceType.GetGenericTypeDefinition() == typeof(Nullable<>)))
                {
                    method.Statements.Add(
                        new CodeConditionStatement(
                            new CodeBinaryOperatorExpression(
                                new CodePropertyReferenceExpression(
                                    new CodePropertyReferenceExpression(
                                        new CodeVariableReferenceExpression("xmlData"),
                                        field.Name
                                    ),
                                    "HasValue"
                                ),
                                CodeBinaryOperatorType.IdentityEquality,
                                new CodePrimitiveExpression(true)
                            ),
                            new CodeStatement[] {
                                    new CodeExpressionStatement(
                                        codeExpression
                                    )
                                }
                        ));
                }
                else if (field.InstanceType == typeof(string))
                {
                    method.Statements.Add(
                        new CodeConditionStatement(
                            new CodeBinaryOperatorExpression(
                                new CodePropertyReferenceExpression(
                                    new CodeVariableReferenceExpression("xmlData"),
                                    field.Name
                                ),
                                CodeBinaryOperatorType.IdentityInequality,
                                new CodePrimitiveExpression(null)
                            ),
                            new CodeStatement[] {
                                    new CodeExpressionStatement(
                                        codeExpression
                                    )
                                }
                        ));
                }
                else
                {
                    method.Statements.Add(codeExpression);
                }                
            }


            method.Statements.Add(new CodeCommentStatement("Done with properties"));


            method.Statements.Add(new CodeVariableDeclarationStatement(
                    typeof(IData),
                    "newData",
                    new CodeObjectCreateExpression(
                        _wrapperClassName,
                        new CodeExpression[] {
                            new CodeVariableReferenceExpression("newElement"),
                            new CodeObjectCreateExpression(
                                typeof(DataSourceId),
                                new CodeExpression[] {
                                    new CodeObjectCreateExpression(
                                        _dataIdClassName,
                                        new CodeVariableReferenceExpression("newElement")
                                    ),
                                    new CodeVariableReferenceExpression("providerName"),
                                    new CodeTypeOfExpression(InterfaceName)
                                }
                            )
                        }
                    )
                ));


            method.Statements.Add(new CodeMethodReturnStatement(
                    new CodeCastExpression(
                        new CodeTypeReference(genericParameter),
                        new CodeVariableReferenceExpression("newData")
                    )
                ));


            declaration.Members.Add(method);
        }


        private string InterfaceFullName { get { return TypeManager.GetRuntimeFullName(_dataTypeDescriptor.TypeManagerTypeName); } }

        
        private string InterfaceName
        {
            get
            {
                if (_interfaceName == null)
                {
                    _interfaceName = InterfaceFullName;
                    if (_interfaceName.IndexOf(',') >= 0)
                    {
                        _interfaceName = _interfaceName.Remove(_interfaceName.IndexOf(','));
                    }
                }

                return _interfaceName;
            }
        }
    }
}
