/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.Text;
using Composite.C1Console.Security;
using Composite.Core.IO;
using Composite.Core.Serialization;


namespace Composite.Plugins.Elements.ElementProviders.WebsiteFileElementProvider
{
    /// <summary>    
    /// </summary>
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)] 
    [SecurityAncestorProvider(typeof(WebsiteFileProviderEntityTokenSecurityAncestorProvider))]
    public sealed class WebsiteFileElementProviderEntityToken : EntityToken
    {
        private static readonly string _baseDirectory;
        private readonly string _providerName;
        private readonly string _relativePath;

        static WebsiteFileElementProviderEntityToken()
        {
            string baseDirectory = PathUtil.BaseDirectory;
            if(baseDirectory.EndsWith("\\"))
            {
                baseDirectory = baseDirectory.Substring(0, baseDirectory.Length - 1);
            }

            _baseDirectory = baseDirectory;
        }

        /// <exclude />
        public WebsiteFileElementProviderEntityToken(string providerName, string path, string rootPath)
        {
            Verify.ArgumentNotNull(path, "path");
            Verify.ArgumentNotNull(rootPath, "rootPath");
            Verify.ArgumentCondition(path.StartsWith(rootPath, StringComparison.OrdinalIgnoreCase), "path", "Path should start with root path");

            Verify.That(rootPath.StartsWith(_baseDirectory, StringComparison.OrdinalIgnoreCase), "Path does not belong to the website");

            _relativePath = path.Substring(_baseDirectory.Length);

            this.Path = path;
            this.RootPath = rootPath;
            _providerName = providerName;
        }


        /// <exclude />
        public string ProviderName
        {
            get { return _providerName; }
        }


        /// <exclude />
        public override string Type
        {
            get { return ""; }
        }


        /// <exclude />
        public override string Source
        {
            get { return _providerName; }
        }


        /// <exclude />
        public override string Id
        {
            get { return _relativePath; }
        }


        /// <exclude />
        public override string Serialize()
        {
            StringBuilder sb = new StringBuilder();

            DoSerialize(sb);

            StringConversionServices.SerializeKeyValuePair(sb, "root", RelativeRootPath);
             
            return sb.ToString();
        }


        /// <exclude />
        public static EntityToken Deserialize(string serializedData)
        {
            Dictionary<string, string> dic;
            string type, source, id;

            EntityToken.DoDeserialize(serializedData, out type, out source, out id, out dic);

            // Backward compatibility
            if(dic.ContainsKey("RootPath"))
            {
                string rootPath = StringConversionServices.DeserializeValueString(dic["RootPath"]);

                return new WebsiteFileElementProviderEntityToken(source, id, rootPath);
            }
            
            string relativeRootPath = StringConversionServices.DeserializeValueString(dic["root"]);

            return new WebsiteFileElementProviderEntityToken(source, _baseDirectory + id, _baseDirectory + relativeRootPath);
        }


        /// <exclude />
        public string Path
        {
            get;
            private set;
        }

        private string RelativeRootPath
        {
            get { return RootPath.Substring(_baseDirectory.Length); }
        }

        internal string RootPath
        {
            get;
            private set;
        }


        internal bool IsRoot
        {
            get { return this.Path == this.RootPath; }
        }


        /// <exclude />
        public override bool Equals(object obj)
        {
            return base.Equals(obj) 
                   && (obj as WebsiteFileElementProviderEntityToken).RootPath == this.RootPath;
        }


        /// <exclude />
        public override int GetHashCode()
        {
            if (this.HashCode == 0)
            {
                this.HashCode = GetType().GetHashCode() ^ this.Type.GetHashCode() ^ this.Source.GetHashCode() ^ this.Id.GetHashCode() ^ this.RelativeRootPath.GetHashCode();
            }
            return this.HashCode;
        }
    }
}
