/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Xml.Linq;
using Composite.Core.Extensions;
using Composite.Core.Types;
using Composite.Core.WebClient.Renderings.Page;
using Composite.Core.Xml;
using Composite.Functions;

namespace Composite.Plugins.PageTemplates.MasterPages.Controls.Functions
{
    /// <exclude />
    [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Never)] 
    [PersistChildren(false)]
    [ParseChildren(true, "Parameters")]
    public class Function : Control
    {
        private static readonly string LogTitle = "Controls.Function";

        /// <exclude />
        public string Name { get; set; }

        private ParamCollection _parameters;

        /// <exclude />
        [PersistenceMode(PersistenceMode.InnerDefaultProperty)]
        [DefaultValue(default(string))]
        [MergableProperty(false)]
        public ParamCollection Parameters
        {
            get
            {
                if (_parameters == null)
                {
                    _parameters = new ParamCollection();
                }

                return _parameters;
            }
        }

        /// <exclude />
        protected override void OnInit(EventArgs e)
        {
            Type returnType;
            object result;
            string functionName = null;

            var functionContextContainer = PageRenderer.GetPageRenderFunctionContextContainer();

            try
            {
                functionName = Name;

                IFunction function;

                if (!FunctionFacade.TryGetFunction(out function, functionName))
                {
                    throw new InvalidOperationException("Invalid function name '{0}'".FormatWith(functionName));
                }

                returnType = function.ReturnType;

                IDictionary<string, object> parameters = parseParameters();

                VerifyParameterMatch(function, parameters);

                result = ExecuteFunction(function, parameters, functionContextContainer);
            }
            catch (Exception ex)
            {
                XElement errorBoxHtml;
                if (!functionContextContainer.ProcessException(functionName, ex, LogTitle, out errorBoxHtml))
                {
                    throw;
                }

                result = errorBoxHtml;
                returnType = typeof(XElement);
            }

            if (result != null)
            {
                if (returnType == typeof(XElement) || returnType == typeof(XhtmlDocument))
                {
                    var element = ValueTypeConverter.Convert<XElement>(result);
                    var markup = new Markup(element, functionContextContainer);

                    Controls.Add(markup);
                }
                else if (typeof(Control).IsAssignableFrom(returnType))
                {
                    var control = (Control) result;

                    Controls.Add(control);
                }
                else if (result is IEnumerable<XNode>)
                {
                    var nodes = result as IEnumerable<XNode>;

                    foreach (XNode node in nodes)
                    {
                        if (node == null) continue;

                        if (node is XElement)
                        {
                            var markup = new Markup(node as XElement, functionContextContainer);

                            Controls.Add(markup);
                        }
                        else
                        {
                            Controls.Add(new LiteralControl(node.ToString()));
                        }
                        
                    }
                }
                else if (result is XAttribute)
                {
                    var parentControl = this.Parent as HtmlGenericControl;
                    if(parentControl != null)
                    {
                        var attr = (XAttribute) result;
                        parentControl.Attributes.Add(attr.Name.ToString(), attr.Value);
                    }
                    else
                    {
                        const string comment = @"<!-- Failed to add attribute, parent control should be of type HtmlGenericControl, check that runat=""server"" attribute is added -->";
                        Controls.Add(new LiteralControl(comment));
                    }
                }
                else 
                {
                    var str = result.ToString();

                    Controls.Add(new LiteralControl(str));
                }
            }

            base.OnInit(e);
        }

        /// <summary>
        /// Executes the function. Note that all the XhtmlParameters will have all the nested &gt;f:function /&lt; lazily evaluated
        /// </summary>
        private static object ExecuteFunction(IFunction function, IDictionary<string, object> parameters, FunctionContextContainer functionContextContainer )
        {
            List<BaseParameterRuntimeTreeNode> parameterNodes = new List<BaseParameterRuntimeTreeNode>();

            if (parameters != null)
            {
                foreach (KeyValuePair<string, object> kvp in parameters)
                {
                    var value = kvp.Value;
                    if (value != null && value is XhtmlDocument)
                    {
                        parameterNodes.Add(new LazyParameterRuntimeTreeNode(kvp.Key,
                            () => ExecuteNestedFunctions(value as XhtmlDocument, functionContextContainer)));
                    }
                    else
                    {
                        parameterNodes.Add(new ConstantObjectParameterRuntimeTreeNode(kvp.Key, kvp.Value));
                    }
                }
            }

            var treeNode = new FunctionRuntimeTreeNode(function, parameterNodes);

            return treeNode.GetValue(functionContextContainer);
        }

        private static XhtmlDocument ExecuteNestedFunctions(XhtmlDocument document, FunctionContextContainer functionContextContainer)
        {
            PageRenderer.ExecuteEmbeddedFunctions(document.Root, functionContextContainer);

            return document;
        }

        private static void VerifyParameterMatch(IFunction function, IDictionary<string, object> parameters)
        {
            var initializationInfo = function as IFunctionInitializationInfo;
            if(initializationInfo != null && !initializationInfo.FunctionInitializedCorrectly) return;
            
            foreach(string parameterName in parameters.Keys)
            {
                Verify.That(function.ParameterProfiles.Any(p => p.Name == parameterName),
                    "Function '{0}.{1}' does not have parameter '{2}' defined",
                    function.Namespace, function.Name, parameterName);
            }
        }

        private IDictionary<string, object> parseParameters()
        {
            var result = new Dictionary<string, object>();

            foreach (Param param in Parameters)
            {
                param.DataBind();

                result.Add(param.Name, param.Value);
            }

            return result;
        }
    }
}