/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

ButtonStateManager.STATE_NORMAL	= 0;
ButtonStateManager.STATE_HOVER	= 1;
ButtonStateManager.STATE_ACTIVE	= 2;
ButtonStateManager.RIGHT_BUTTON = 2;

/**
 * @class
 * Better externalize this complex stuff from the ButtonBinding.
 * @param {ButtonBinding} binding
 */
function ButtonStateManager ( binding ) {
	
	/**
	 * @type {SystemLogger}
	 */
	this.logger = SystemLogger.getLogger ( "ButtonStateManager" );
	
	/**
	 * @type {ButtonBinding}
	 */
	this.binding = binding;
	
	/**
	 * @type {ImageProfile}
	 */
	this.imageProfile = binding.imageProfile;

	/* 
	 * Assigning event listener.
	 */
	this.assignDOMEvents(true);

	/*
	* Returnable.
	*/
	return this;
}

/**
 * Assigning DOM event listeners.
 */
ButtonStateManager.prototype.assignDOMEvents = function ( isAssign ) {
	
	var action = isAssign ? "addEventListener" : "removeEventListener";
	
	this.binding [ action ] ( DOMEvents.MOUSEENTER, this );
	this.binding [ action ] ( DOMEvents.MOUSELEAVE, this );
	this.binding [ action ] ( DOMEvents.MOUSEDOWN, this );
	this.binding [ action ] ( DOMEvents.MOUSEUP, this );
}

/**
 * Cleanup.
 * @see {ButtonBinding#onBindingDispose}
 */
ButtonStateManager.prototype.dispose = function () {
	
	this.assignDOMEvents ( false );
	this.binding = null;
	this.imageProfile = null;
}

/** 
 * TODO: Split into multiple methods.
 * @mplements {IEventListener}.
 * @param {MouseEvent} e
 */
ButtonStateManager.prototype.handleEvent = function (e) {

	if (Binding.exists(this.binding) && !this.binding.isDisabled && !BindingDragger.isDragging) {

		var isCommand = false, isPopup = false, state = null;

		if (e.button == ButtonStateManager.RIGHT_BUTTON) {
			// do nothing - right clicks are handled by the contextmenu property
		}
		else if (this.binding.isCheckBox) {

			switch (e.type) {
				case DOMEvents.MOUSEENTER:
				case DOMEvents.MOUSEOVER:
					state = ButtonStateManager.STATE_HOVER; // image decision left to imageprofile!
					break;
				case DOMEvents.MOUSELEAVE:
				case DOMEvents.MOUSEOUT:
					state = this.binding.isChecked ? ButtonStateManager.STATE_ACTIVE : ButtonStateManager.STATE_NORMAL;
					break;
				case DOMEvents.MOUSEDOWN:
					state = ButtonStateManager.STATE_HOVER;
					break;
				case DOMEvents.MOUSEUP:
					this.binding.isChecked = !this.binding.isChecked;
					state = this.binding.isChecked ? ButtonStateManager.STATE_ACTIVE : ButtonStateManager.STATE_NORMAL;
					if (state == ButtonStateManager.STATE_ACTIVE) {
						this.binding._check(true);
					} else {
						this.binding._uncheck(true);
					}
					isCommand = true;
					break;
			}
		}
		else if (this.binding.isComboButton) {
			switch (e.type) {
				case DOMEvents.MOUSEENTER:
				case DOMEvents.MOUSEOVER:
					state = ButtonStateManager.STATE_HOVER;
					break;
				case DOMEvents.MOUSELEAVE:
				case DOMEvents.MOUSEOUT:
					state = ButtonStateManager.STATE_NORMAL;
					break;
				case DOMEvents.MOUSEDOWN:
					state = ButtonStateManager.STATE_ACTIVE;
					break;
				case DOMEvents.MOUSEUP:
					state = ButtonStateManager.STATE_NORMAL;
					var targetBinding = UserInterface.getBinding(e.target ? e.target : e.srcElement);
					if (targetBinding instanceof ComboBoxBinding) {
						this.binding.isChecked = !this.binding.isChecked;
						state = this.binding.isChecked ? ButtonStateManager.STATE_ACTIVE : ButtonStateManager.STATE_NORMAL;
						if (state == ButtonStateManager.STATE_ACTIVE) {
							this.binding._check(true);
						} else {
							this.binding._uncheck(true);
						}
						isPopup = true;
					}
					else {
						if (this.binding.isChecked)
							this.binding._uncheck(true);
						state = ButtonStateManager.STATE_NORMAL;
						isCommand = true;
					}
					break;
			}
		}
		else if (this.binding.isCheckButton || this.binding.isRadioButton) {

			switch (e.type) {
				case DOMEvents.MOUSEENTER:
				case DOMEvents.MOUSEOVER:
					if (!this.binding.isChecked) {
						state = ButtonStateManager.STATE_HOVER;
					}
					break;
				case DOMEvents.MOUSELEAVE:
				case DOMEvents.MOUSEOUT:
					if (!this.binding.isChecked) { // TODO: CHECK DESCENDANT TARGET!
						state = ButtonStateManager.STATE_NORMAL;
					}
					break;
				case DOMEvents.MOUSEDOWN:
					state = ButtonStateManager.STATE_ACTIVE;
					break;
				case DOMEvents.MOUSEUP:
					if (this.binding.isCheckButton || !this.binding.isChecked) {
						this.binding.isChecked = !this.binding.isChecked;
						state = this.binding.isChecked ? ButtonStateManager.STATE_ACTIVE : ButtonStateManager.STATE_NORMAL;
						if (state == ButtonStateManager.STATE_ACTIVE) {
							this.binding._check(true);
						} else {
							this.binding._uncheck(true);
						}
						isCommand = true;
					}
					break;
			}
		}
		else {
			switch (e.type) {
				case DOMEvents.MOUSEENTER:
				case DOMEvents.MOUSEOVER:
					state = ButtonStateManager.STATE_HOVER;
					break;
				case DOMEvents.MOUSELEAVE:
				case DOMEvents.MOUSEOUT:
					state = ButtonStateManager.STATE_NORMAL;
					break;
				case DOMEvents.MOUSEDOWN:
					state = ButtonStateManager.STATE_ACTIVE;
					break;
				case DOMEvents.MOUSEUP:
					state = ButtonStateManager.STATE_NORMAL;
					isCommand = true;
					break;
			}
		}
		switch (state) {
			case ButtonStateManager.STATE_NORMAL:
				this.invokeNormalState();
				break;
			case ButtonStateManager.STATE_HOVER:
				this.invokeHoverState();
				break;
			case ButtonStateManager.STATE_ACTIVE:
				this.invokeActiveState();
				break;
		}

		/*
		* Please note that the button command is invoked here!
		*/
		if (isCommand) {
			this.binding.fireCommand();
		}

		if (isPopup) {
			this.binding.invokePopup();
		}

		/*
		* Check patches explorer in case fireCommand 
		* closed the containing window.
		*/
		if (Binding.exists(this.binding) == true) {

			/*
			* Consuming the event!
			*/
			DOMEvents.stopPropagation(e);

			/*
			* Broadcast mousedown and mouseup to close open popups and menupopups. 
			* Notice that our binding is broadcasted as argument. This will prevent 
			* popups associated to *our* binding from closing as soon as it opens. 
			* The broadcast argument should really be a MouseEvent, so this is 
			* really a terrible way to handle the popup display status problem. 
			* UPDATE: we now use Bindings as arguments in other scenarios.
			* @see {PopupBinding#handleBroadcast}
			*/
			switch (e.type) {
				case DOMEvents.MOUSEDOWN:
					this.binding.onMouseDown();
					break;
				case DOMEvents.MOUSEUP:
					this.binding.onMouseUp();
					break;
			}
		}
	}
}

/**
 * Invoke normal state.
 */
ButtonStateManager.prototype.invokeNormalState = function () {

	this.binding.detachClassName ( "hover" );
	this.binding.detachClassName ( "active" );
	this.binding.detachClassName ( "isdisabled" );
	
	if ( this.imageProfile ) {
		var url = this.imageProfile.getDefaultImage ();
		if ( url ) {
			this.binding.setImage ( url );
		}
	}
}

/**
 * Invoke hover state.
 */
ButtonStateManager.prototype.invokeHoverState = function () {

	this.binding.attachClassName ( "hover" );
	this.binding.detachClassName ( "active" );
	
	if ( this.imageProfile ) {
		var url = this.imageProfile.getHoverImage ();
		if ( url ) {
			this.binding.setImage ( url );
		}
	}
}

/**
 * Invoke active state.
 */
ButtonStateManager.prototype.invokeActiveState = function () {

	this.binding.attachClassName ( "active" );
	this.binding.detachClassName ( "hover" );
	
	if ( this.imageProfile ) {
		var url = this.imageProfile.getActiveImage ();
		if ( url ) {
			this.binding.setImage ( url );
		}
	}
}

/**
 * Invoke disabled state. This method gets invoked by the button.
 * @see {ButtonBinding#setDisabled}
 */
ButtonStateManager.prototype.invokeDisabledState = function () {

	this.binding.detachClassName ( "hover" );
	this.binding.detachClassName ( "active" );
	this.binding.attachClassName ( "isdisabled" );
	
	if ( this.imageProfile ) {
		var image = this.imageProfile.getDisabledImage ();
		if ( image ) {
			this.binding.setImage ( image );
		}
	}
}