/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

CursorBinding.prototype = new Binding;
CursorBinding.prototype.constructor = CursorBinding;
CursorBinding.superclass = Binding.prototype;

/**
 * Fade in cursor.
 * @param {CursorBinding} cursor
 */
CursorBinding.fadeIn = function ( cursor ) {
	
	if ( cursor instanceof CursorBinding ) {
		cursor.setOpacity ( 0 );
		cursor.show ();
		new Animation (
			{
				modifier : 9,
				onstep : function ( iterator ) {
					cursor.setOpacity ( 
						Math.sin ( iterator * Math.PI / 180 )
					);
				},
				onstop : function () {
					cursor.setOpacity ( 1 );
				}
			}
		).play ();
	}
}

/**
 * Fade out cursor.
 * @param {CursorBinding} cursor
 */
CursorBinding.fadeOut = function ( cursor ) {
	
	if ( cursor instanceof CursorBinding ) {
		new Animation ({
			modifier : 9,
			onstep : function ( iterator ) {
				cursor.setOpacity ( 
					Math.cos ( 
						iterator * Math.PI / 180
					)
				);
			},
			onstop : function () {
				cursor.hide ();
			}
		}).play ();
	}
}

/**
 * Move cursor from one point to another whilst fading it out.
 * @param {CursorBinding} cursor
 */
CursorBinding.moveOut = function ( cursor, startpoint, endpoint ) {
	
	if ( cursor instanceof CursorBinding ) {
		
		// compensate for cursor icons visual displacement
		endpoint.x -= 16;
		endpoint.y -= 16;
	
		new Animation ({
			modifier : 3,
			onstep : function ( iterator ) {
				var tal = Math.sin ( iterator * Math.PI / 180 );
				cursor.setPosition ( new Point (
					(( 1 - tal ) * startpoint.x ) + (( 0 + tal ) * endpoint.x ),
					(( 1 - tal ) * startpoint.y ) + (( 0 + tal ) * endpoint.y )
				));
			},
			onstop : function () {
				CursorBinding.fadeOut ( cursor )
			}
		}).play ();
	}
}

/**
 * @class
 */
function CursorBinding () {

	/**
	 * @type {SystemLogger}
	 */
	this.logger = SystemLogger.getLogger ( "CursorBinding" );
	
	/**
	 * @type {LabelBinding}
	 */
	this._labelBinding = null;
	
	/**
	 * @type {number}
	 */
	this._opacity = 1;
	
	/**
	 * @type {boolean}
	 */
	this.isAccepting = true;
	
	/*
	 * @returnable.
	 */
	return this;
}

/**
 * Identifies binding.
 */
CursorBinding.prototype.toString = function () {
	
	return "[CursorBinding]";
}

/**
 * @overloads {Binding#onBindingAttach}
 */
CursorBinding.prototype.onBindingAttach = function () {
	
	CursorBinding.superclass.onBindingAttach.call ( this );
	
	this._labelBinding = this.add (
		LabelBinding.newInstance ( 
			this.bindingDocument 
		)
	);
	
	var image = this.getProperty ( "image" );
	if ( image != null ) {
		this.setImage ( image );
	}
	
	this._stopIndicatorBinding = this.add (
		LabelBinding.newInstance ( 
			this.bindingDocument 
		)
	);
	
	this._stopIndicatorBinding.attachClassName ( "indicator" );
	this._stopIndicatorBinding.setImage ( 
		"${icon:cancel}" 
	);
	
	this.hide ();
	this._stopIndicatorBinding.hide ();
}

/**
 * @param {string} url
 */
CursorBinding.prototype.setImage = function ( url ) {
	
	this._labelBinding.setImage ( url );
}

/**
 * Show acceptance.
 */
CursorBinding.prototype.showAcceptance = function () {
	
	this.isAccepting = true;
	
	if ( Client.isMozilla ) {
		this._stopIndicatorBinding.hide ();
	} else {
		var self = this;
		setTimeout ( function () {
			if ( self.isAccepting ) {
				self._stopIndicatorBinding.hide ();
			}
		}, 0 );
	}
}

/**
 * Hide acceptance.
 */
CursorBinding.prototype.hideAcceptance = function () {

	this.isAccepting = false;
	
	if ( Client.isMozilla ) {
		this._stopIndicatorBinding.show ();
	} else {
		var self = this;
		setTimeout ( function () {
			if ( !self.isAccepting ) {
				self._stopIndicatorBinding.show ();
			}
		}, 0 );
	}
}

/**
 * @overloads {Binding#show}
 */
CursorBinding.prototype.show = function () {

	CursorBinding.superclass.show.call ( this );
	
	/*
	this._stopIndicatorBinding.show ()
	this._isAccepting = false;
	*/
}

/**
 * @param {number} opacity From zero to one!
 */
CursorBinding.prototype.setOpacity = function ( opacity ) {

	this.bindingElement.style.opacity = new String ( opacity );

	this._opacity = opacity;
}

/**
 * @return {number}
 */
CursorBinding.prototype.getOpacity = function () {
	
	return this._opacity;
}

/**
 * @param {Position} pos
 */
CursorBinding.prototype.setPosition = function ( pos ) {
	
	this.bindingElement.style.left = pos.x + "px";
	this.bindingElement.style.top = pos.y + "px";
}

/**
 * @return {Position}
 */
CursorBinding.prototype.getPosition = function () {
	
	return new Point (
		this.bindingElement.offsetLeft,
		this.bindingElement.offsetTop
	);
}

/**
 * Fade in.
 */
CursorBinding.prototype.fadeIn = function () {
	
	CursorBinding.fadeIn ( this ); 
}

/**
 * Fade out.
 */
CursorBinding.prototype.fadeOut = function () {
	
	CursorBinding.fadeOut ( this ); 
}