/*
 * The contents of this web application are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this web application except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 * 
 * The Original Code is owned by and the Initial Developer of the Original Code is 
 * Composite A/S (Danish business reg.no. 21744409). All Rights Reserved
 * 
 * Section 11 of the License is EXPRESSLY amended to include a provision stating 
 * that any dispute, including but not limited to disputes related to the enforcement 
 * of the License, to which Composite A/S as owner of the Original Code, as Initial 
 * Developer or in any other role, becomes a part to shall be governed by Danish law 
 * and be initiated before the Copenhagen City Court ("K�benhavns Byret")            
 */

using System;
using System.IO;
using System.Reflection;
using System.Text;
using System.Web;
using System.Web.Hosting;
using System.Web.UI;
using Composite.Core.Instrumentation;
using Composite.Core.WebClient;
using Composite.Core.WebClient.Renderings;


public partial class Renderers_Page : System.Web.UI.Page
{
    private IDisposable _pageEventsPageMeasuring;

    private bool _requestCompleted = false;

    RenderingContext _renderingContext;

    protected override void OnPreInit(EventArgs e)
    {
        // If there's no parameters to the request, redirecting to the application root url
        if (Request.RawUrl.Equals( UrlUtils.ResolvePublicUrl("~/Renderers/Page.aspx"), StringComparison.OrdinalIgnoreCase))
        {
            Response.Redirect(HostingEnvironment.ApplicationVirtualPath, true);
            return;
        }

        _renderingContext = RenderingContext.InitializeFromHttpContext();

        InitializeCulture();

        base.OnPreInit(e);
    }

    protected override void OnInit(EventArgs e)
    {
        if (_renderingContext.RunResponseHandlers())
        {
            _requestCompleted = true;
            return;
        }

        using (Profiler.Measure("ASP.NET controls : OnInit"))
        {
            base.OnInit(e);
        }

        _pageEventsPageMeasuring = Profiler.Measure("ASP.NET controls: PageLoad, Event handling, PreRender");
    }



    protected override void Render(HtmlTextWriter writer)
    {
        if (_requestCompleted)
        {
            return;
        }

        if (_pageEventsPageMeasuring != null)
        {
            _pageEventsPageMeasuring.Dispose();
        }

        ScriptManager scriptManager = ScriptManager.GetCurrent(this);
        bool isUpdatePanelPostback = scriptManager != null && scriptManager.IsInAsyncPostBack;

        if (isUpdatePanelPostback == true)
        {
            base.Render(writer);
            return;
        }

        _renderingContext.PreRenderRedirectCheck();

        var markupBuilder = new StringBuilder();
        var sw = new StringWriter(markupBuilder);
        try
        {
            using (Profiler.Measure("ASP.NET controls: Render"))
            {
                base.Render(new HtmlTextWriter(sw));
            }
        }
        catch (HttpException ex)
        {
            CheckForMultipleAspNetFormsException(ex);
            throw;
        }

        string xhtml = _renderingContext.ConvertInternalLinks(markupBuilder.ToString());

        xhtml = _renderingContext.FormatXhtml(xhtml);

        // Inserting perfomance profiling information
        if (_renderingContext.ProfilingEnabled)
        {
            xhtml = _renderingContext.BuildProfilerReport();

            Response.ContentType = "text/xml";
        }

        writer.Write(xhtml);
    }


    private void CheckForMultipleAspNetFormsException(HttpException ex)
    {
        MethodInfo setStringMethod = typeof(HttpContext).Assembly /* System.Web */
                .GetType("System.Web.SR")
                .GetMethod("GetString", BindingFlags.Public | BindingFlags.Static, null, new[] { typeof(string) }, null);

        string multipleFormNotAllowedMessage = (string)setStringMethod.Invoke(null, new object[] { "Multiple_forms_not_allowed" });

        bool multipleAspFormTagsExists = ex.Message == multipleFormNotAllowedMessage;
        if (multipleAspFormTagsExists)
        {
            string errorMessage = (this.MasterPageFile != null)
                ? "Multiple <form runat=\"server\" /> elements exist on this page." +
                  "ASP.NET supports only one visible <form runat=\"server\" /> control at a time.\n " +
                  "To fix this, insert a <form runat=\"server\"> ... </form> tag in your master page template that spans all placeholders."
                : "Multiple <asp:form /> elements exist on this page. " +
                  "ASP.NET supports only one visible <form runat=\"server\" /> control at a time.\n " +
                  "To fix this, insert a <asp:form> ... </asp:form> section in your template that spans all controls.";
            
            throw new HttpException(errorMessage);

        }
    }

    protected override void OnUnload(EventArgs e)
    {
        base.OnUnload(e);

        if (_renderingContext != null)
        {
            _renderingContext.Dispose();
        }
    }


    protected override void InitializeCulture()
    {
        if (_renderingContext != null)
        {
            this.Culture = this.UICulture = _renderingContext.Page.DataSourceId.LocaleScope.Name;
        }

        base.InitializeCulture();
    }
}